# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Utility functions to help in building responses from ORM models"""
from typing import Any, Callable, Final, Optional, TypedDict, cast
from urllib.parse import urlencode, urlunparse

from marshmallow import Schema

from dioptra.restapi.db import models
from dioptra.restapi.routes import V1_ROOT

ARTIFACTS: Final[str] = "artifacts"
ENTRYPOINTS: Final[str] = "entrypoints"
EXPERIMENTS: Final[str] = "experiments"
GROUPS: Final[str] = "groups"
MODELS: Final[str] = "models"
PLUGINS: Final[str] = "plugins"
PLUGIN_FILES: Final[str] = "files"
PLUGIN_PARAMETER_TYPES: Final[str] = "pluginParameterTypes"
QUEUES: Final[str] = "queues"
TAGS: Final[str] = "tags"
USERS: Final[str] = "users"

# -- Typed Dictionaries --------------------------------------------------------


class GroupRefDict(TypedDict):
    id: int
    name: str
    url: str


class PluginParameterTypeRefDict(TypedDict):
    id: int
    group: GroupRefDict
    url: str
    name: str


class PluginTaskInputParameterDict(TypedDict):
    name: str
    required: bool
    parameter_type: PluginParameterTypeRefDict


class PluginTaskOutputParameterDict(TypedDict):
    name: str
    parameter_type: PluginParameterTypeRefDict


class PluginTaskDict(TypedDict):
    name: str
    input_params: list[PluginTaskInputParameterDict]
    output_params: list[PluginTaskOutputParameterDict]


class PluginWithFilesDict(TypedDict):
    plugin: models.Plugin
    plugin_files: list[models.PluginFile]
    has_draft: bool | None


class PluginFileDict(TypedDict):
    plugin_file: models.PluginFile
    plugin: models.Plugin
    has_draft: bool | None


class ExperimentDict(TypedDict):
    experiment: models.Experiment
    entrypoints: list[models.EntryPoint]
    queue: models.Queue | None
    has_draft: bool | None


class PluginParameterTypeDict(TypedDict):
    plugin_task_parameter_type: models.PluginTaskParameterType
    has_draft: bool | None


class ArtifactDict(TypedDict):
    artifact: models.Artifact
    has_draft: bool | None


class QueueDict(TypedDict):
    queue: models.Queue
    has_draft: bool | None


class EntrypointDict(TypedDict):
    entry_point: models.EntryPoint
    queues: list[models.Queue]
    has_draft: bool | None


class JobDict(TypedDict):
    job: models.Job
    artifacts: list[models.Artifact]
    has_draft: bool | None


class ModelWithVersionDict(TypedDict):
    ml_model: models.MlModel
    version: models.MlModelVersion | None
    has_draft: bool | None


# -- Ref Types -----------------------------------------------------------------


def build_experiment_ref(experiment: models.Experiment) -> dict[str, Any]:
    """Build an ExperimentRef dictionary.

    Args:
        experiment: The experiment object to convert into an ExperimentRef dictionary.

    Returns:
        The ExperimentRef dictionary.
    """
    return {
        "id": experiment.resource_id,
        "name": experiment.name,
        "group": build_group_ref(experiment.resource.owner),
        "url": build_url(f"{EXPERIMENTS}/{experiment.resource_id}"),
    }


def build_experiment_snapshot_ref(experiment: models.Experiment) -> dict[str, Any]:
    """Build an ExperimentSnapshotRef dictionary.

    Args:
        experiment: The experiment object to convert into an ExperimentSnapshotRef
            dictionary.

    Returns:
        The ExperimentSnapshotRef dictionary.
    """
    return {
        "snapshot_id": experiment.resource_snapshot_id,
        "name": experiment.name,
        "group": build_group_ref(experiment.resource.owner),
        "url": build_url(
            f"{EXPERIMENTS}/{experiment.resource_id}/snapshots/"
            f"{experiment.resource_snapshot_id}"
        ),
    }


def build_user_ref(user: models.User) -> dict[str, Any]:
    """Build a UserRef dictionary.

    Args:
        user: The User object to convert into a UserRef dictionary.

    Returns:
        The UserRef dictionary.
    """
    return {
        "id": user.user_id,
        "username": user.username,
        "url": build_url(f"{USERS}/{user.user_id}"),
    }


def build_group_ref(group: models.Group) -> GroupRefDict:
    """Build a GroupRef dictionary.

    Args:
        group: The Group object to convert into a GroupRef dictionary.

    Returns:
        The GroupRef dictionary.
    """
    return {
        "id": group.group_id,
        "name": group.name,
        "url": build_url(f"{GROUPS}/{group.group_id}"),
    }


def build_tag_ref(tag: models.Tag) -> dict[str, Any]:
    """Build a TagRef dictionary.

    Args:
        tag: The Tag object to convert into a TagRef dictionary.

    Returns:
        The TagRef dictionary.
    """
    return {
        "id": tag.tag_id,
        "name": tag.name,
        "group": build_group_ref(tag.owner),
        "url": build_url(f"{TAGS}/{tag.tag_id}"),
    }


def build_tag_ref_list(tags: list[models.Tag]) -> list[dict[str, Any]]:
    """Build a TagRef list.

    Args:
        tag: The list of Tag objects to convert into a TagRef list.

    Returns:
        The TagRef list.
    """
    return [build_tag_ref(tag) for tag in tags]


def build_plugin_ref(plugin: models.Plugin) -> dict[str, Any]:
    """Build a PluginRef dictionary.

    Args:
        plugin: The Plugin object to convert into a PluginRef dictionary.

    Returns:
        The PluginRef dictionary.
    """
    return {
        "id": plugin.resource_id,
        "name": plugin.name,
        "group": build_group_ref(plugin.resource.owner),
        "url": build_url(f"{PLUGINS}/{plugin.resource_id}"),
    }


def build_entrypoint_plugin(plugin_with_files: PluginWithFilesDict) -> dict[str, Any]:
    """Build a PluginFileRef dictionary.

    Args:
        entrypoint: The Entrypoint object
        plugin: The Plugin object to convert into a PluginRef dictionary.

    Returns:
        The PluginFileRef dictionary.
    """
    plugin = plugin_with_files["plugin"]
    plugin_files = plugin_with_files["plugin_files"]

    return {
        "id": plugin.resource_id,
        "snapshot_id": plugin.resource_snapshot_id,
        "name": plugin.name,
        "url": build_url(
            f"{PLUGINS}/{plugin.resource_id}/snapshots/{plugin.resource_snapshot_id}"
        ),
        "files": [
            {
                "id": plugin_file.resource_id,
                "snapshot_id": plugin_file.resource_snapshot_id,
                "filename": plugin_file.filename,
                "url": build_url(
                    f"{PLUGINS}/{plugin.resource_id}/{PLUGIN_FILES}/{plugin_file.resource_id}/"
                    f"snapshots/{plugin_file.resource_snapshot_id}"
                ),
            }
            for plugin_file in plugin_files
        ],
    }


def build_plugin_snapshot_ref(plugin: models.Plugin) -> dict[str, Any]:
    """Build a PluginSnapshotRef dictionary.

    Args:
        plugin: The PluginFile object to convert into a PluginSnapshotRef dictionary.

    Returns:
        The PluginSnapshotRef dictionary.
    """
    return {
        "snapshot_id": plugin.resource_snapshot_id,
        "name": plugin.name,
        "group": build_group_ref(plugin.resource.owner),
        "url": build_url(
            f"{PLUGINS}/{plugin.resource_id}/snapshots/{plugin.resource_snapshot_id}"
        ),
    }


def build_plugin_file_ref(plugin_file: models.PluginFile) -> dict[str, Any]:
    """Build a PluginRef dictionary.

    Args:
        queue: The Plugin object to convert into a PluginRef dictionary.

    Returns:
        The PluginRef dictionary.
    """
    plugin_id = plugin_file.plugin_id
    return {
        "id": plugin_file.resource_id,
        "plugin": plugin_id,
        "filename": plugin_file.filename,
        "group": build_group_ref(plugin_file.resource.owner),
        "url": build_url(
            f"{PLUGINS}/{plugin_id}/{PLUGIN_FILES}/{plugin_file.resource_id}"
        ),
    }


def build_entrypoint_ref(entrypoint: models.EntryPoint) -> dict[str, Any]:
    """Build a EntrypointRef dictionary.

    Args:
        entrypoint: The Entrypoint object to convert into a EntrypointRef dictionary.

    Returns:
        The EntrypointRef dictionary.
    """
    return {
        "id": entrypoint.resource_id,
        "name": entrypoint.name,
        "group": build_group_ref(entrypoint.resource.owner),
        "url": build_url(f"{ENTRYPOINTS}/{entrypoint.resource_id}"),
    }


def build_entrypoint_snapshot_ref(entrypoint: models.EntryPoint) -> dict[str, Any]:
    """Build a EntrypointSnapshotRef dictionary.

    Args:
        queue: The Entrypoint object to convert into a EntrypointSnapshotRef dictionary.

    Returns:
        The EntrypointSnapshotRef dictionary.
    """
    return {
        "snapshot_id": entrypoint.resource_snapshot_id,
        "name": entrypoint.name,
        "group": build_group_ref(entrypoint.resource.owner),
        "url": build_url(
            f"{ENTRYPOINTS}/{entrypoint.resource_id}"
            f"/snapshots/{entrypoint.resource_snapshot_id}"
        ),
    }


def build_model_ref(model: models.MlModel) -> dict[str, Any]:
    """Build a MlModel dictionary.

    Args:
        model: The Model object to convert into a MlModel dictionary.

    Returns:
        The MlModel dictionary.
    """
    return {
        "id": model.resource_id,
        "name": model.name,
        "group": build_group_ref(model.resource.owner),
        "url": f"{MODELS}/{model.resource_id}",
    }


def build_queue_ref(queue: models.Queue) -> dict[str, Any]:
    """Build a QueueRef dictionary.

    Args:
        queue: The Queue object to convert into a QueueRef dictionary.

    Returns:
        The QueueRef dictionary.
    """
    return {
        "id": queue.resource_id,
        "name": queue.name,
        "group": build_group_ref(queue.resource.owner),
        "url": build_url(f"{QUEUES}/{queue.resource_id}"),
    }


def build_queue_snapshot_ref(queue: models.Queue) -> dict[str, Any]:
    """Build a QueueSnapshotRef dictionary.

    Args:
        queue: The Queue object to convert into a QueueSnapshotRef dictionary.

    Returns:
        The QueueSnapshotRef dictionary.
    """
    return {
        "snapshot_id": queue.resource_snapshot_id,
        "name": queue.name,
        "group": build_group_ref(queue.resource.owner),
        "url": build_url(
            f"{QUEUES}/{queue.resource_id}/snapshots/{queue.resource_snapshot_id}"
        ),
    }


def build_plugin_parameter_type_ref(
    plugin_param_type: models.PluginTaskParameterType,
) -> PluginParameterTypeRefDict:
    """Build a PluginParameterTypeRef dictionary.

    Args:
        plugin_param_type: The Plugin Parameter Type object to convert into a
            PluginParameterTypeRef dictionary.

    Returns:
        The PluginParameterTypeRef dictionary.
    """
    return {
        "id": plugin_param_type.resource_id,
        "name": plugin_param_type.name,
        "group": build_group_ref(plugin_param_type.resource.owner),
        "url": build_url(f"{PLUGIN_PARAMETER_TYPES}/{plugin_param_type.resource_id}"),
    }


def build_artifact_ref(artifact: models.Artifact) -> dict[str, Any]:
    """Build a ArtifactRef dictionary.

    Args:
        artifact: The Artifact object to convert into a ArtifactRef dictionary.

    Returns:
        The ArtifactRef dictionary.
    """
    return {
        "id": artifact.resource_id,
        "artifact_uri": artifact.uri,
        "group": build_group_ref(artifact.resource.owner),
        "url": build_url(f"{ARTIFACTS}/{artifact.resource_id}"),
    }


# -- Full Types ----------------------------------------------------------------


def build_user(user: models.User) -> dict[str, Any]:
    """Build a User response dictionary.

    Args:
        user: The User object to convert into a User response dictionary.

    Returns:
        The User response dictionary.
    """
    return {
        "id": user.user_id,
        "username": user.username,
        "email": user.email_address,
    }


def build_current_user(user: models.User) -> dict[str, Any]:
    """Build a response dictionary for the current user.

    Args:
        user: The User object representing the current user to convert into a response
            dictionary.

    Returns:
        The response dictionary for the current user.
    """
    return {
        "id": user.user_id,
        "username": user.username,
        "email": user.email_address,
        "groups": [
            build_group_ref(membership.group) for membership in user.group_memberships
        ],
        "created_on": user.created_on,
        "last_modified_on": user.last_modified_on,
        "last_login_on": user.last_login_on,
        "password_expires_on": user.password_expire_on,
    }


def build_group(group: models.Group) -> dict[str, Any]:
    """Build a Group response dictionary.

    Args:
        group: The Group object to convert into a Group response dictionary.

    Returns:
        The Group response dictionary.
    """
    members: dict[int, dict[str, Any]] = {}

    for member in group.members:
        members[member.user_id] = {
            "user": build_user_ref(member.user),
            "group": build_group_ref(group),
            "permissions": {
                "read": member.read,
                "write": member.write,
                "share_read": member.share_read,
                "share_write": member.share_write,
                "owner": False,
                "admin": False,
            },
        }

    for manager in group.managers:
        members[manager.user_id]["permissions"]["owner"] = manager.owner
        members[manager.user_id]["permissions"]["admin"] = manager.admin

    return {
        "id": group.group_id,
        "name": group.name,
        "user": build_user_ref(group.creator),
        "members": list(members.values()),
        "created_on": group.created_on,
        "last_modified_on": group.last_modified_on,
    }


def build_experiment(experiment_dict: ExperimentDict) -> dict[str, Any]:
    """Build an Experiment response dictionary.

    Args:
        experiment: The experiment object to convert into an Experiment response
            dictionary.

    Returns:
        The Experiment response dictionary.
    """
    experiment = experiment_dict["experiment"]
    entrypoints = experiment_dict.get("entrypoints", None)
    has_draft = experiment_dict.get("has_draft", None)

    data = {
        "id": experiment.resource_id,
        "snapshot_id": experiment.resource_snapshot_id,
        "name": experiment.name,
        "description": experiment.description,
        "user": build_user_ref(experiment.creator),
        "group": build_group_ref(experiment.resource.owner),
        "created_on": experiment.resource.created_on,
        "last_modified_on": experiment.resource.last_modified_on,
        "latest_snapshot": experiment.resource.latest_snapshot_id
        == experiment.resource_snapshot_id,
        "tags": [build_tag_ref(tag) for tag in experiment.tags],
    }

    if entrypoints is not None:
        data["entrypoints"] = [
            build_entrypoint_ref(entrypoint) for entrypoint in entrypoints
        ]

    if has_draft is not None:
        data["has_draft"] = has_draft

    return data


def build_tag(tag: models.Tag) -> dict[str, Any]:
    """Build a Tag response dictionary.

    Args:
        queue: The Tag object to convert into a Tag response dictionary.

    Returns:
        The Tag response dictionary.
    """
    return {
        "id": tag.tag_id,
        "name": tag.name,
        "user": build_user_ref(tag.creator),
        "group": build_group_ref(tag.owner),
        "created_on": tag.created_on,
        "last_modified_on": tag.last_modified_on,
    }


def build_entrypoint(entrypoint_dict: EntrypointDict) -> dict[str, Any]:
    entrypoint = entrypoint_dict["entry_point"]
    queues = entrypoint_dict.get("queues", None)
    has_draft = entrypoint_dict.get("has_draft", None)

    plugins_dict = {
        entry_point_plugin_file.plugin.resource_id: PluginWithFilesDict(
            plugin=entry_point_plugin_file.plugin, plugin_files=[], has_draft=False
        )
        for entry_point_plugin_file in entrypoint.entry_point_plugin_files
    }
    for entry_point_plugin_file in entrypoint.entry_point_plugin_files:
        resource_id = entry_point_plugin_file.plugin.resource_id
        plugin_file = entry_point_plugin_file.plugin_file
        plugins_dict[resource_id]["plugin_files"].append(plugin_file)
    plugins = list(plugins_dict.values())

    data = {
        "id": entrypoint.resource_id,
        "snapshot_id": entrypoint.resource_snapshot_id,
        "name": entrypoint.name,
        "description": entrypoint.description,
        "task_graph": entrypoint.task_graph,
        "user": build_user_ref(entrypoint.creator),
        "group": build_group_ref(entrypoint.resource.owner),
        "created_on": entrypoint.resource.created_on,
        "last_modified_on": entrypoint.resource.last_modified_on,
        "latest_snapshot": entrypoint.resource.latest_snapshot_id
        == entrypoint.resource_snapshot_id,
        "tags": [build_tag_ref(tag) for tag in entrypoint.tags],
        "parameters": [
            {
                "name": param.name,
                "default_value": param.default_value,
                "parameter_type": param.parameter_type,
            }
            for param in entrypoint.parameters
        ],
        "plugins": [build_entrypoint_plugin(plugin) for plugin in plugins],
    }

    if queues is not None:
        data["queues"] = [build_queue_ref(queue) for queue in queues]

    if has_draft is not None:
        data["has_draft"] = has_draft

    return data


def build_job(job_dict: JobDict) -> dict[str, Any]:
    """Build a Job response dictionary.

    Args:
        job: The Job object to convert into a job response dictionary.

    Returns:
        The Job response dictionary.
    """
    job = job_dict["job"]
    artifacts = job_dict.get("artifacts", None)
    has_draft = job_dict.get("has_draft", None)

    data = {
        "id": job.resource_id,
        "snapshot_id": job.resource_snapshot_id,
        "description": job.description,
        "status": job.status,
        "values": {
            param.parameter.name: param.value
            for param in job.entry_point_job.entry_point_parameter_values
        },
        "timeout": job.timeout,
        "user": build_user_ref(job.creator),
        "group": build_group_ref(job.resource.owner),
        "queue": build_queue_snapshot_ref(job.queue_job.queue),
        "experiment": build_experiment_snapshot_ref(job.experiment_job.experiment),
        "entrypoint": build_entrypoint_snapshot_ref(job.entry_point_job.entry_point),
        "created_on": job.resource.created_on,
        "last_modified_on": job.resource.last_modified_on,
        "latest_snapshot": job.resource.latest_snapshot_id == job.resource_snapshot_id,
        "tags": [build_tag_ref(tag) for tag in job.tags],
    }

    if artifacts is not None:
        data["artifacts"] = [build_artifact_ref(artifact) for artifact in artifacts]

    if has_draft is not None:
        data["has_draft"] = has_draft

    return data


def build_model(model_dict: ModelWithVersionDict) -> dict[str, Any]:
    """Build a Model response dictionary for the latest version.

    Args:
        model: The Model object to convert into a Model response dictionary.

    Returns:
        The Model response dictionary.
    """
    model = model_dict["ml_model"]
    version = model_dict.get("version", None)
    has_draft = model_dict.get("has_draft", None)

    latest_version = build_model_version(model_dict) if version is not None else None
    latest_version_number = version.version_number if version is not None else 0

    # WARNING: this assumes versions cannot be deleted and that no details of
    # the version # are needed in constructing the ref type. If either of these
    # assumptions change, the service and controller layers will need to be
    # updated to return the MlModelVersion # ORM objects.
    versions = [
        {
            "version_number": version_number,
            "url": build_url(f"{MODELS}/{model.resource_id}/versions/{version_number}"),
        }
        for version_number in range(1, latest_version_number + 1)
    ]

    data = {
        "id": model.resource_id,
        "snapshot_id": model.resource_snapshot_id,
        "name": model.name,
        "description": model.description,
        "user": build_user_ref(model.creator),
        "group": build_group_ref(model.resource.owner),
        "created_on": model.resource.created_on,
        "last_modified_on": model.resource.last_modified_on,
        "latest_snapshot": model.resource.latest_snapshot_id
        == model.resource_snapshot_id,
        "tags": [build_tag_ref(tag) for tag in model.tags],
        "latest_version": latest_version,
        "versions": versions,
    }

    if has_draft is not None:
        data["has_draft"] = has_draft

    return data


def build_model_version(model_dict: ModelWithVersionDict) -> dict[str, Any]:
    """Build a ModelVersion response dictionary.

    Args:
        model: The ModelVersion object to convert into a ModelVersion response
            dictionary.

    Returns:
        The ModelVersion response dictionary.
    """
    model = model_dict["ml_model"]
    version = cast(models.MlModelVersion, model_dict["version"])
    return {
        "model": build_model_ref(model),
        "description": version.description,
        "version_number": version.version_number,
        "artifact": build_artifact_ref(version.artifact),
        "created_on": version.resource.created_on,
    }


def build_artifact(artifact_dict: ArtifactDict) -> dict[str, Any]:
    artifact = artifact_dict["artifact"]
    has_draft = artifact_dict.get("has_draft")

    data = {
        "id": artifact.resource_id,
        "snapshot_id": artifact.resource_snapshot_id,
        "uri": artifact.uri,
        "description": artifact.description,
        "user": build_user_ref(artifact.creator),
        "group": build_group_ref(artifact.resource.owner),
        "created_on": artifact.resource.created_on,
        "last_modified_on": artifact.resource.last_modified_on,
        "latest_snapshot": artifact.resource.latest_snapshot_id
        == artifact.resource_snapshot_id,
        "tags": [build_tag_ref(tag) for tag in artifact.tags],
    }

    if has_draft is not None:
        data["has_draft"] = has_draft

    return data


def build_queue(queue_dict: QueueDict) -> dict[str, Any]:
    """Build a Queue response dictionary.

    Args:
        queue: The Queue object to convert into a Queue response dictionary.

    Returns:
        The Queue response dictionary.
    """
    queue = queue_dict["queue"]
    has_draft = queue_dict.get("has_draft", None)

    data = {
        "id": queue.resource_id,
        "snapshot_id": queue.resource_snapshot_id,
        "name": queue.name,
        "description": queue.description,
        "user": build_user_ref(queue.creator),
        "group": build_group_ref(queue.resource.owner),
        "created_on": queue.resource.created_on,
        "last_modified_on": queue.resource.last_modified_on,
        "latest_snapshot": queue.resource.latest_snapshot_id
        == queue.resource_snapshot_id,
        "tags": [build_tag_ref(tag) for tag in queue.tags],
    }

    if has_draft is not None:
        data["has_draft"] = has_draft

    return data


def build_plugin(plugin_with_files: PluginWithFilesDict) -> dict[str, Any]:
    """Build a Plugin response dictionary.

    Args:
        queue: The Plugin object to convert into a Plugin response dictionary.

    Returns:
        The Plugin response dictionary.
    """
    plugin = plugin_with_files["plugin"]
    plugin_files = plugin_with_files.get("plugin_files", None)
    has_draft = plugin_with_files.get("has_draft", None)

    data = {
        "id": plugin.resource_id,
        "snapshot_id": plugin.resource_snapshot_id,
        "name": plugin.name,
        "description": plugin.description,
        "user": build_user_ref(plugin.creator),
        "group": build_group_ref(plugin.resource.owner),
        "created_on": plugin.resource.created_on,
        "last_modified_on": plugin.resource.last_modified_on,
        "latest_snapshot": plugin.resource.latest_snapshot_id
        == plugin.resource_snapshot_id,
        "tags": [build_tag_ref(tag) for tag in plugin.tags],
    }

    if plugin_files is not None:
        data["files"] = [
            build_plugin_file_ref(plugin_file) for plugin_file in plugin_files
        ]

    if has_draft is not None:
        data["has_draft"] = has_draft

    return data


def build_plugin_file(plugin_file_with_plugin: PluginFileDict) -> dict[str, Any]:
    plugin_file = plugin_file_with_plugin["plugin_file"]
    plugin = plugin_file_with_plugin.get("plugin", None)
    has_draft = plugin_file_with_plugin.get("has_draft", None)

    data = {
        "id": plugin_file.resource_id,
        "snapshot_id": plugin_file.resource_snapshot_id,
        "filename": plugin_file.filename,
        "description": plugin_file.description,
        "user": build_user_ref(plugin_file.creator),
        "group": build_group_ref(plugin_file.resource.owner),
        "created_on": plugin_file.resource.created_on,
        "last_modified_on": plugin_file.resource.last_modified_on,
        "latest_snapshot": plugin_file.resource.latest_snapshot_id
        == plugin_file.resource_snapshot_id,
        "contents": plugin_file.contents,
        "tasks": [build_plugin_task(task) for task in plugin_file.tasks],
        "tags": [build_tag_ref(tag) for tag in plugin_file.tags],
    }

    if plugin is not None:
        data["plugin"] = build_plugin_ref(plugin)

    if has_draft is not None:
        data["has_draft"] = has_draft

    return data


def build_plugin_task(plugin_task: models.PluginTask) -> PluginTaskDict:
    input_params: list[PluginTaskInputParameterDict] = []
    for input_parameter in sorted(
        plugin_task.input_parameters, key=lambda x: x.parameter_number
    ):
        input_params.append(
            PluginTaskInputParameterDict(
                name=input_parameter.name,
                required=input_parameter.required,
                parameter_type=build_plugin_parameter_type_ref(
                    input_parameter.parameter_type
                ),
            )
        )

    output_params: list[PluginTaskOutputParameterDict] = []
    for output_parameter in sorted(
        plugin_task.output_parameters, key=lambda x: x.parameter_number
    ):
        output_params.append(
            PluginTaskOutputParameterDict(
                name=output_parameter.name,
                parameter_type=build_plugin_parameter_type_ref(
                    output_parameter.parameter_type
                ),
            )
        )

    return PluginTaskDict(
        name=plugin_task.plugin_task_name,
        input_params=input_params,
        output_params=output_params,
    )


def build_plugin_parameter_type(
    plugin_parameter_type_dict: PluginParameterTypeDict,
) -> dict[str, Any]:
    """Build a Plugin Parameter Type response dictionary.

    Args:
        plugin_parameter_type: The Plugin Parameter Type object to convert
            into a Plugin Parameter Type response dictionary.

    Returns:
        The Plugin Parameter Type response dictionary.
    """
    plugin_parameter_type = plugin_parameter_type_dict["plugin_task_parameter_type"]
    has_draft = plugin_parameter_type_dict["has_draft"]

    data = {
        "id": plugin_parameter_type.resource_id,
        "snapshot_id": plugin_parameter_type.resource_snapshot_id,
        "name": plugin_parameter_type.name,
        "structure": plugin_parameter_type.structure,
        "description": plugin_parameter_type.description,
        "user": build_user_ref(plugin_parameter_type.creator),
        "group": build_group_ref(plugin_parameter_type.resource.owner),
        "created_on": plugin_parameter_type.resource.created_on,
        "last_modified_on": plugin_parameter_type.resource.last_modified_on,
        "latest_snapshot": plugin_parameter_type.resource.latest_snapshot_id
        == plugin_parameter_type.resource_snapshot_id,
        "tags": [build_tag_ref(tag) for tag in plugin_parameter_type.tags],
    }

    if has_draft is not None:
        data["has_draft"] = has_draft

    return data


def build_resource_draft(
    draft: models.DraftResource,
    draft_schema: type[Schema] | Schema,
    num_other_drafts: int | None = None,
) -> dict[str, Any]:
    """Build a Draft response dictionary for a resource.

    Args:
        draft: The Draft object to convert into a Draft response dictionary.

    Returns:
        The Draft response dictionary.
    """

    if isinstance(draft_schema, Schema):
        schema = draft_schema
    else:
        schema = draft_schema()
    payload = schema.dump(draft.payload["resource_data"])

    metadata = dict()
    if num_other_drafts is not None:
        metadata["num_other_drafts"] = num_other_drafts
    return {
        "id": draft.draft_resource_id,
        "resource_id": draft.payload.get("resource_id", None),
        "resource_snapshot_id": draft.payload.get("resource_snapshot_id", None),
        "payload": payload,
        "resource_type": draft.resource_type,
        "user": build_user_ref(draft.creator),
        "group": build_group_ref(draft.target_owner),
        "created_on": draft.created_on,
        "last_modified_on": draft.last_modified_on,
        "metadata": metadata,
    }


def build_resource_url(resource: models.Resource):
    return build_url(f"{resource.resource_type}/{resource.resource_id}")


# -- Paging --------------------------------------------------------------------


def build_paging_envelope(
    route_prefix: str,
    build_fn: Callable,
    data: list[Any],
    group_id: int | None,
    query: str | None,
    draft_type: str | None,
    index: int,
    length: int,
    total_num_elements: int,
    sort_by: Optional[str] = None,
    descending: Optional[bool] = None,
) -> dict[str, Any]:
    """Build the paging envelope for a response.

    Args:
        route_prefix: The prefix of the route, forms the URL path in the paging url.
        build_fn: The function for converting an ORM object into a response dictionary.
            This dictionary is then wrapped in the paging envelope and set as the "data"
            field.
        data: The list of ORM objects to wrap in the paging envelope.
        query: The optional search query string.
        draft_type: The type of drafts to return.
        index: The index of the current page.
        length: The number of results to return per page.
        total_num_elements: The total number of elements in the collection.
        sort_by: The name of the column to sort.
        descending: Boolean indicating whether to sort by descending or not.

    Returns:
        The paging envelope for the response.
    """
    has_prev = index > 0
    has_next = total_num_elements > index + length
    is_complete = not has_next

    paged_data = {
        "index": index,
        "is_complete": is_complete,
        "total_num_results": total_num_elements,
        "sort_by": sort_by,
        "descending": descending,
        "first": build_paging_url(
            route_prefix,
            group_id=group_id,
            search=query,
            draft_type=draft_type,
            index=0,
            length=length,
        ),
        "data": [build_fn(x) for x in data],
    }

    if has_prev:
        prev_index = max(index - length, 0)
        prev_url = build_paging_url(
            route_prefix,
            group_id=group_id,
            search=query,
            draft_type=draft_type,
            index=prev_index,
            length=length,
        )
        paged_data["prev"] = prev_url

    if has_next:
        next_index = index + length
        next_url = build_paging_url(
            route_prefix,
            group_id=group_id,
            search=query,
            draft_type=draft_type,
            index=next_index,
            length=length,
        )
        paged_data["next"] = next_url

    return paged_data


def build_paging_url(
    route_prefix: str,
    group_id: int | None,
    search: str | None,
    draft_type: str | None,
    index: int,
    length: int,
) -> str:
    """Build a URL for a paged resource endpoint.

    Args:
        resource_type: The prefix of the route to paginate, forms the URL path.
        search: The optional search query string.
        draft_type: The type of drafts to return.
        index: The index of the current page.
        length: The number of results to return per page.

    Returns:
        A quoted URL string for the paged resource endpoint.
    """
    query_params: dict[str, Any] = {"index": index, "pageLength": length}

    if group_id:
        query_params["groupId"] = group_id

    if search:
        query_params["search"] = search

    if draft_type:
        query_params["draft_type"] = draft_type

    return build_url(route_prefix, query_params)


def build_url(route_prefix: str, query_params: dict[str, str] | None = None) -> str:
    query_params = query_params or {}

    return urlunparse(
        ("", "", f"/{V1_ROOT}/{route_prefix}", "", urlencode(query_params), "")
    )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from typing import Tuple

import structlog
from structlog.stdlib import BoundLogger

LOGGER: BoundLogger = structlog.stdlib.get_logger()


try:
    from tensorflow.keras.layers import (
        Activation,
        BatchNormalization,
        Conv2D,
        Layer,
        LeakyReLU,
    )

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )

DEFAULT_ACTIVATION = LeakyReLU(alpha=0.1)


class Conv2DBatchNormalization(Layer):
    def __init__(
        self,
        filters: int,
        kernel_size: Tuple[int, int],
        strides: Tuple[int, int] = (1, 1),
        padding: str = "same",
        data_format=None,
        dilation_rate: Tuple[int, int] = (1, 1),
        groups: int = 1,
        activation=DEFAULT_ACTIVATION,
        use_bias: bool = True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        batch_norm_before_activation=True,
        name="conv",
        **kwargs,
    ):
        super().__init__()
        self.conv2d = Conv2D(
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs,
        )
        self.batch_norm = BatchNormalization(name=f"{name}_batch_norm")
        self.activation = Activation(activation, name=f"{name}_activation")
        self.batch_norm_before_activation = batch_norm_before_activation
        self._batch_norm_and_activation_fn = (
            self._batch_norm_before_activation
            if batch_norm_before_activation
            else self._batch_norm_after_activation
        )

    def call(self, inputs, training=None):
        x = self.conv2d(inputs)

        return self._batch_norm_and_activation_fn(x=x, training=training)

    def get_config(self):
        conv2d_arg_names = [
            "filters",
            "kernel_size",
            "strides",
            "padding",
            "data_format",
            "dilation_rate",
            "groups",
            "activation",
            "use_bias",
            "kernel_initializer",
            "bias_initializer",
            "kernel_regularizer",
            "bias_regularizer",
            "activity_regularizer",
            "kernel_constraint",
            "bias_constraint",
            "name",
        ]
        conv2d_config = {
            k: v for k, v in self.conv2d.get_config().items() if k in conv2d_arg_names
        }

        config = super().get_config()
        config.update(conv2d_config)
        config["batch_norm_before_activation"] = self.batch_norm_before_activation

        return config

    def _batch_norm_after_activation(self, x, training):
        x = self.activation(x)
        return self.batch_norm(x, training=training)

    def _batch_norm_before_activation(self, x, training):
        x = self.batch_norm(x, training=training)
        return self.activation(x)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The schemas for serializing/deserializing Entrypoint resources."""
from marshmallow import Schema, fields, validate

from dioptra.restapi.v1.queues.schema import QueueRefSchema
from dioptra.restapi.v1.schemas import (
    BasePageSchema,
    GroupIdQueryParametersSchema,
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
    generate_base_resource_ref_schema,
    generate_base_resource_schema,
)


class EntrypointPluginFileSchema(Schema):
    """The schema for the data stored in a Entrypoint PluginFile snapshot."""

    id = fields.Integer(
        attribute="id",
        metadata=dict(description="ID for the PluginFile resource."),
    )
    filename = fields.String(
        attribute="filename",
        metadata=dict(description="Filename of the PluginFile resource."),
    )
    snapshotId = fields.Integer(
        attribute="snapshot_id",
        metadata=dict(description="Snapshot ID for the PluginFile resource."),
    )
    url = fields.Url(
        attribute="url",
        metadata=dict(description="URL for accessing the full PluginFile snapshot."),
        relative=True,
    )


class EntrypointPluginSchema(Schema):
    """The schema for the data stored in a Entrypoint Plugin snapshot."""

    id = fields.Integer(
        attribute="id",
        metadata=dict(description="ID for the Plugin resource."),
    )
    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Plugin resource."),
    )
    snapshotId = fields.Integer(
        attribute="snapshot_id",
        metadata=dict(description="Snapshot ID for the Plugin resource."),
    )
    url = fields.Url(
        attribute="url",
        metadata=dict(description="URL for accessing the full Plugin snapshot."),
        relative=True,
    )
    files = fields.Nested(
        EntrypointPluginFileSchema,
        attribute="files",
        many=True,
        metadata=dict(description="List of parameters for the entrypoint."),
    )


class EntrypointParameterSchema(Schema):
    """The schema for the data stored in a Entrypoint parameter resource."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Entrypoint parameter resource."),
        required=True,
    )
    defaultValue = fields.String(
        attribute="default_value",
        metadata=dict(description="Default value of the Entrypoint parameter."),
        load_default=None,
    )
    parameterType = fields.String(
        attribute="parameter_type",
        metadata=dict(description="Data type of the Entrypoint parameter."),
        required=True,
        validate=validate.OneOf(
            ["string", "float", "integer", "boolean", "list", "mapping"]
        ),
    )


EntrypointRefBaseSchema = generate_base_resource_ref_schema("Entrypoint")
EntrypointSnapshotRefBaseSchema = generate_base_resource_ref_schema(
    "Entrypoint", keep_snapshot_id=True
)


class EntrypointRefSchema(EntrypointRefBaseSchema):  # type: ignore
    """The reference schema for the data stored in a Entrypoint resource."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Entrypoint resource."),
    )


class EntrypointSnapshotRefSchema(EntrypointSnapshotRefBaseSchema):  # type: ignore
    """The snapshot reference schema for the data stored in a Entrypoint resource."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Entrypoint resource."),
    )


class EntrypointMutableFieldsSchema(Schema):
    """The fields schema for the mutable data in a Entrypoint resource."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Entrypoint resource."),
        required=True,
    )
    description = fields.String(
        attribute="description",
        metadata=dict(description="Description of the Entrypoint resource."),
        load_default=None,
    )
    taskGraph = fields.String(
        attribute="task_graph",
        metadata=dict(description="Task graph of the Entrypoint resource."),
        required=True,
    )
    parameters = fields.Nested(
        EntrypointParameterSchema,
        attribute="parameters",
        many=True,
        metadata=dict(description="List of parameters for the entrypoint."),
    )
    queueIds = fields.List(
        fields.Integer(),
        attribute="queue_ids",
        data_key="queues",
        metadata=dict(description="The queue for the entrypoint."),
        load_only=True,
    )


class EntrypointPluginMutableFieldsSchema(Schema):
    pluginIds = fields.List(
        fields.Integer(),
        attribute="plugin_ids",
        data_key="plugins",
        metadata=dict(description="List of plugin files for the entrypoint."),
        load_only=True,
    )


EntrypointBaseSchema = generate_base_resource_schema("Entrypoint", snapshot=True)


class EntrypointSchema(
    EntrypointPluginMutableFieldsSchema,
    EntrypointMutableFieldsSchema,
    EntrypointBaseSchema,  # type: ignore
):
    """The schema for the data stored in a Entrypoint resource."""

    plugins = fields.Nested(
        EntrypointPluginSchema,
        attribute="plugins",
        many=True,
        metadata=dict(description="List of plugins for the entrypoint."),
        dump_only=True,
    )
    queues = fields.Nested(
        QueueRefSchema,
        attribute="queues",
        many=True,
        metadata=dict(description="The queue for the entrypoint."),
        dump_only=True,
    )


class EntrypointDraftSchema(
    EntrypointPluginMutableFieldsSchema,
    EntrypointMutableFieldsSchema,
    EntrypointBaseSchema,  # type: ignore
):
    """The schema for the data stored in a Entrypoint resource."""

    pluginIds = fields.List(
        fields.Integer(),
        attribute="plugin_ids",
        data_key="plugins",
        metadata=dict(description="List of plugin files for the entrypoint."),
    )
    queueIds = fields.List(
        fields.Integer(),
        attribute="queue_ids",
        data_key="queues",
        metadata=dict(description="The queue for the entrypoint."),
    )


class EntrypointPageSchema(BasePageSchema):
    """The paged schema for the data stored in a Entrypoint resource."""

    data = fields.Nested(
        EntrypointSchema,
        many=True,
        metadata=dict(description="List of Entrypoint resources in the current page."),
    )


class EntrypointGetQueryParameters(
    PagingQueryParametersSchema,
    GroupIdQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
):
    """The query parameters for the GET method of the /entrypoints endpoint."""



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for Entrypoint resources."""
from __future__ import annotations

import uuid
from typing import cast

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import inject
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.restapi.routes import V1_ENTRYPOINTS_ROUTE
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.queues.schema import QueueRefSchema
from dioptra.restapi.v1.schemas import IdListSchema, IdStatusResponseSchema
from dioptra.restapi.v1.shared.drafts.controller import (
    generate_resource_drafts_endpoint,
    generate_resource_drafts_id_endpoint,
    generate_resource_id_draft_endpoint,
)
from dioptra.restapi.v1.shared.snapshots.controller import (
    generate_resource_snapshots_endpoint,
    generate_resource_snapshots_id_endpoint,
)
from dioptra.restapi.v1.shared.tags.controller import (
    generate_resource_tags_endpoint,
    generate_resource_tags_id_endpoint,
)

from .schema import (
    EntrypointDraftSchema,
    EntrypointGetQueryParameters,
    EntrypointMutableFieldsSchema,
    EntrypointPageSchema,
    EntrypointPluginMutableFieldsSchema,
    EntrypointPluginSchema,
    EntrypointSchema,
)
from .service import (
    RESOURCE_TYPE,
    SEARCHABLE_FIELDS,
    EntrypointIdPluginsIdService,
    EntrypointIdPluginsService,
    EntrypointIdQueuesIdService,
    EntrypointIdQueuesService,
    EntrypointIdService,
    EntrypointService,
)

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace("Entrypoints", description="Entrypoints endpoint")


@api.route("/")
class EntrypointEndpoint(Resource):
    @inject
    def __init__(self, entrypoint_service: EntrypointService, *args, **kwargs) -> None:
        """Initialize the entrypoint resource.

        All arguments are provided via dependency injection.

        Args:
            entrypoint_service: A EntrypointService object.
        """
        self._entrypoint_service = entrypoint_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=EntrypointGetQueryParameters, api=api)
    @responds(schema=EntrypointPageSchema, api=api)
    def get(self):
        """Gets a list of Entrypoint resources matching the filter parameters."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Entrypoint", request_type="GET"
        )

        parsed_query_params = request.parsed_query_params  # noqa: F841
        group_id = parsed_query_params["group_id"]
        search_string = parsed_query_params["search"]
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]
        sort_by_string = parsed_query_params["sort_by"]
        descending = parsed_query_params["descending"]

        entrypoints, total_num_entrypoints = self._entrypoint_service.get(
            group_id=group_id,
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
            sort_by_string=sort_by_string,
            descending=descending,
            log=log,
        )
        return utils.build_paging_envelope(
            "entrypoints",
            build_fn=utils.build_entrypoint,
            data=entrypoints,
            group_id=group_id,
            draft_type=None,
            query=search_string,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_entrypoints,
            sort_by=sort_by_string,
            descending=descending,
        )

    @login_required
    @accepts(schema=EntrypointSchema, api=api)
    @responds(schema=EntrypointSchema, api=api)
    def post(self):
        """Creates an Entrypoint resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Entrypoint", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # noqa: F841
        entrypoint = self._entrypoint_service.create(
            name=parsed_obj["name"],
            description=parsed_obj["description"],
            task_graph=parsed_obj["task_graph"],
            parameters=parsed_obj["parameters"],
            plugin_ids=parsed_obj["plugin_ids"],
            queue_ids=parsed_obj["queue_ids"],
            group_id=int(parsed_obj["group_id"]),
            log=log,
        )
        return utils.build_entrypoint(entrypoint)


@api.route("/<int:id>")
@api.param("id", "ID for the Entrypoint resource.")
class EntrypointIdEndpoint(Resource):
    @inject
    def __init__(
        self, entrypoint_id_service: EntrypointIdService, *args, **kwargs
    ) -> None:
        """Initialize the entrypoint resource.

        All arguments are provided via dependency injection.

        Args:
            entrypoint_id_service: A EntrypointIdService object.
        """
        self._entrypoint_id_service = entrypoint_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=EntrypointSchema, api=api)
    def get(self, id: int):
        """Gets an Entrypoint resource by its unique ID."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="Entrypoint",
            request_type="GET",
            id=id,
        )
        entrypoint = cast(
            models.EntryPoint,
            self._entrypoint_id_service.get(id, error_if_not_found=True, log=log),
        )
        return utils.build_entrypoint(entrypoint)

    @login_required
    @accepts(schema=EntrypointMutableFieldsSchema, api=api)
    @responds(schema=EntrypointSchema, api=api)
    def put(self, id: int):
        """Modifies an Entrypoint resource by its unique ID."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="Entrypoint",
            request_type="PUT",
            id=id,
        )
        parsed_obj = request.parsed_obj  # type: ignore # noqa: F841
        entrypoint = cast(
            models.EntryPoint,
            self._entrypoint_id_service.modify(
                id,
                name=parsed_obj["name"],
                description=parsed_obj["description"],
                task_graph=parsed_obj["task_graph"],
                parameters=parsed_obj["parameters"],
                queue_ids=parsed_obj["queue_ids"],
                error_if_not_found=True,
                log=log,
            ),
        )
        return utils.build_entrypoint(entrypoint)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int):
        """Deletes an Entrypoint resource by its unique ID."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="Entrypoint",
            request_type="DELETE",
            id=id,
        )
        return self._entrypoint_id_service.delete(entrypoint_id=id, log=log)


@api.route("/<int:id>/plugins")
@api.param("id", "ID for the Entrypoint resource.")
class EntrypointIdPluginsEndpoint(Resource):
    @inject
    def __init__(
        self, entrypoint_id_plugins_service: EntrypointIdPluginsService, *args, **kwargs
    ) -> None:
        """Initialize the entrypoint resource.

        All arguments are provided via dependency injection.

        Args:
            entrypoint_id_plugins_service: A EntrypointIdPluginsService object.
        """
        self._entrypoint_id_plugins_service = entrypoint_id_plugins_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=EntrypointPluginSchema(many=True), api=api)
    def get(self, id: int):
        """Retrieves the plugin snapshots for an entrypoint resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Entrypoint", request_type="POST"
        )
        plugins = self._entrypoint_id_plugins_service.get(id, log=log)
        return [utils.build_entrypoint_plugin(plugin) for plugin in plugins]

    @login_required
    @accepts(schema=EntrypointPluginMutableFieldsSchema, api=api)
    @responds(schema=EntrypointPluginSchema(many=True), api=api)
    def post(self, id: int):
        """Appends plugins to an entrypoint resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Entrypoint", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # type: ignore
        plugins = self._entrypoint_id_plugins_service.append(
            id, plugin_ids=parsed_obj["plugin_ids"], log=log
        )
        return [utils.build_entrypoint_plugin(plugin) for plugin in plugins]


@api.route("/<int:id>/plugins/<int:pluginId>")
@api.param("id", "ID for the Entrypoint resource.")
@api.param("pluginId", "ID for the Plugin resource.")
class EntrypointIdPluginsIdEndpoint(Resource):
    @inject
    def __init__(
        self,
        entrypoint_id_plugins_id_service: EntrypointIdPluginsIdService,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the entrypoint resource.

        All arguments are provided via dependency injection.

        Args:
            entrypoint_id_plugins_id_service: A EntrypointIdPluginsService object.
        """
        self._entrypoint_id_plugins_id_service = entrypoint_id_plugins_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=EntrypointPluginSchema, api=api)
    def get(self, id: int, pluginId: int):
        """Retrieves a plugin snapshot for an entrypoint resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Entrypoint", request_type="POST"
        )
        plugin = self._entrypoint_id_plugins_id_service.get(id, pluginId, log=log)
        return utils.build_entrypoint_plugin(plugin)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int, pluginId: int):
        """Removes a plugin from an entrypoint by ID."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="Entrypoint",
            request_type="DELETE",
            id=id,
        )
        return self._entrypoint_id_plugins_id_service.delete(id, pluginId, log=log)


@api.route("/<int:id>/queues")
@api.param("id", "ID for the Entrypoint resource.")
class EntrypointIdQueuesEndpoint(Resource):
    @inject
    def __init__(
        self, entrypoint_id_queues_service: EntrypointIdQueuesService, *args, **kwargs
    ) -> None:
        self._entrypoint_id_queues_service = entrypoint_id_queues_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=QueueRefSchema(many=True), api=api)
    def get(self, id: int):
        """Gets the list of Queues for the resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Entrypoint", request_type="GET"
        )
        queues = self._entrypoint_id_queues_service.get(id, log=log)
        return [utils.build_queue_ref(queue) for queue in queues]

    @login_required
    @accepts(schema=IdListSchema, api=api)
    @responds(schema=QueueRefSchema(many=True), api=api)
    def post(self, id: int):
        """Appends one or more Queues to the resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Entrypoint", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # type: ignore
        queues = cast(
            list[models.Queue],
            self._entrypoint_id_queues_service.append(
                id, queue_ids=parsed_obj["ids"], error_if_not_found=True, log=log
            ),
        )
        return [utils.build_queue_ref(queue) for queue in queues]

    @login_required
    @accepts(schema=IdListSchema, api=api)
    @responds(schema=QueueRefSchema(many=True), api=api)
    def put(self, id: int):
        """Replaces one or more Queues to the resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Entrypoint", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # type: ignore
        queues = self._entrypoint_id_queues_service.modify(
            id, queue_ids=parsed_obj["ids"], error_if_not_found=True, log=log
        )
        return [utils.build_queue_ref(queue) for queue in queues]

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int):
        """Removes all Queues from the resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Entrypoint", request_type="DELETE"
        )
        return self._entrypoint_id_queues_service.delete(
            id, error_if_not_found=True, log=log
        )


@api.route("/<int:id>/queues/<int:queueId>")
@api.param("id", "ID for the Entrypoint resource.")
@api.param("queueId", "ID for the Queue resource.")
class EntrypointIdQueuesId(Resource):
    @inject
    def __init__(
        self,
        entrypoint_id_queues_id_service: EntrypointIdQueuesIdService,
        *args,
        **kwargs,
    ) -> None:
        self._entrypoint_id_queues_id_service = entrypoint_id_queues_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int, queueId):
        """Removes a Queue from the Entrypoint resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Entrypoint", request_type="GET"
        )
        return self._entrypoint_id_queues_id_service.delete(id, queueId, log=log)


EntrypointDraftResource = generate_resource_drafts_endpoint(
    api,
    resource_name=RESOURCE_TYPE,
    route_prefix=V1_ENTRYPOINTS_ROUTE,
    request_schema=EntrypointDraftSchema,
)
EntrypointDraftIdResource = generate_resource_drafts_id_endpoint(
    api,
    resource_name=RESOURCE_TYPE,
    request_schema=EntrypointDraftSchema(exclude=["groupId"]),
)
EntrypointIdDraftResource = generate_resource_id_draft_endpoint(
    api,
    resource_name=RESOURCE_TYPE,
    request_schema=EntrypointDraftSchema(exclude=["groupId"]),
)

EntrypointSnapshotsResource = generate_resource_snapshots_endpoint(
    api=api,
    resource_model=models.EntryPoint,
    resource_name=RESOURCE_TYPE,
    route_prefix=V1_ENTRYPOINTS_ROUTE,
    searchable_fields=SEARCHABLE_FIELDS,
    page_schema=EntrypointPageSchema,
    build_fn=utils.build_entrypoint,
)
EntrypointSnapshotsIdResource = generate_resource_snapshots_id_endpoint(
    api=api,
    resource_model=models.EntryPoint,
    resource_name=RESOURCE_TYPE,
    response_schema=EntrypointSchema,
    build_fn=utils.build_entrypoint,
)

EntrypointTagsResource = generate_resource_tags_endpoint(
    api=api,
    resource_name=RESOURCE_TYPE,
)
EntrypointTagsIdResource = generate_resource_tags_id_endpoint(
    api=api,
    resource_name=RESOURCE_TYPE,
)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for Tag resources."""
from __future__ import annotations

import uuid
from typing import cast
from urllib.parse import unquote

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import inject
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.schemas import IdStatusResponseSchema, ResourceUrlsPageSchema

from .schema import (
    TagGetQueryParameters,
    TagMutableFieldsSchema,
    TagPageSchema,
    TagResourceQueryParameters,
    TagSchema,
)
from .service import TagIdResourcesService, TagIdService, TagService

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace("Tags", description="Tags endpoint")


@api.route("/")
class TagEndpoint(Resource):
    @inject
    def __init__(self, tag_service: TagService, *args, **kwargs) -> None:
        """Initialize the tag resource.

        All arguments are provided via dependency injection.

        Args:
            tag_service: A TagService object.
        """
        self._tag_service = tag_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=TagGetQueryParameters, api=api)
    @responds(schema=TagPageSchema, api=api)
    def get(self):
        """Gets a list of all Tags."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Tag", request_type="GET"
        )
        parsed_query_params = request.parsed_query_params  # noqa: F841

        group_id = parsed_query_params["group_id"]
        search_string = unquote(parsed_query_params["search"])
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]
        sort_by_string = parsed_query_params["sort_by"]
        descending = parsed_query_params["descending"]

        tags, total_num_tags = self._tag_service.get(
            group_id=group_id,
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
            sort_by_string=sort_by_string,
            descending=descending,
            log=log,
        )
        return utils.build_paging_envelope(
            "tags",
            build_fn=utils.build_tag,
            data=tags,
            group_id=group_id,
            query=search_string,
            draft_type=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_tags,
            sort_by=sort_by_string,
            descending=descending,
        )

    @login_required
    @accepts(schema=TagSchema, api=api)
    @responds(schema=TagSchema, api=api)
    def post(self):
        """Creates a Tag."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Tag", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # noqa: F841

        tag = self._tag_service.create(
            name=str(parsed_obj["name"]),
            group_id=int(parsed_obj["group_id"]),
            log=log,
        )
        return utils.build_tag(tag)


@api.route("/<int:id>")
@api.param("id", "ID for the Tag.")
class TagIdEndpoint(Resource):
    @inject
    def __init__(self, tag_id_service: TagIdService, *args, **kwargs) -> None:
        """Initialize the tag resource.

        All arguments are provided via dependency injection.

        Args:
            tag_id_service: A TagIdService object.
        """
        self._tag_id_service = tag_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=TagSchema, api=api)
    def get(self, id: int):
        """Gets a Tag."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Tag", request_type="GET", id=id
        )
        tag = cast(
            models.Tag,
            self._tag_id_service.get(id, error_if_not_found=True, log=log),
        )
        return utils.build_tag(tag)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int):
        """Deletes a Tag."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Tag", request_type="DELETE", id=id
        )
        return self._tag_id_service.delete(tag_id=id, error_if_not_found=True, log=log)

    @login_required
    @accepts(schema=TagMutableFieldsSchema, api=api)
    @responds(schema=TagSchema, api=api)
    def put(self, id: int):
        """Modifies a Tag."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Tag", request_type="PUT", id=id
        )
        log.debug("Request received")
        parsed_obj = request.parsed_obj  # type: ignore
        tag = cast(
            models.Tag,
            self._tag_id_service.modify(
                id,
                name=parsed_obj["name"],
                error_if_not_found=True,
                log=log,
            ),
        )
        return utils.build_tag(tag)


@api.route("/<int:id>/resources/")
@api.param("id", "ID for the Tag.")
class TagIdResourceEndpoint(Resource):
    @inject
    def __init__(
        self, tag_id_resources_service: TagIdResourcesService, *args, **kwargs
    ) -> None:
        """Initialize the tag resource.

        All arguments are provided via dependency injection.

        Args:
            tag_id_resources_service: A TagService object.
        """
        self._tag_id_resources_service = tag_id_resources_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=TagResourceQueryParameters, api=api)
    @responds(schema=ResourceUrlsPageSchema, api=api)
    def get(self, id: int):
        """Gets all Resources labeled with this Tag."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Tag", request_type="GET", id=id
        )
        parsed_query_params = request.parsed_query_params  # type: ignore # noqa: F841

        resource_type = parsed_query_params["resource_type"]
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]

        resources, total_num_resources = cast(
            tuple[list[models.Resource], int],
            self._tag_id_resources_service.get(
                id,
                resource_type=resource_type,
                page_index=page_index,
                page_length=page_length,
                error_if_not_found=True,
                log=log,
            ),
        )
        return utils.build_paging_envelope(
            f"tags/{id}/resources",
            build_fn=utils.build_resource_url,
            data=resources,
            group_id=None,
            query=None,
            draft_type=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_resources,
        )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import numpy as np
import pytest


@pytest.mark.parametrize(
    "seed",
    [-200, 400, 42, 0],
)
@pytest.mark.parametrize(
    ("low", "high"),
    [(1, None), (20, None), (200, 204), (-200, 10), (-40, -30), (0, 20), (-20, 0)],
)
def test_draw_random_integer(seed, low, high) -> None:
    from dioptra_builtins.random.sample import draw_random_integer

    rng = np.random.default_rng(seed if seed >= 0 else None)
    if seed < 0:
        seed = rng.bit_generator._seed_seq.entropy

    result: int = draw_random_integer(rng, low, high)

    if high is not None:
        assert low <= result and result < high
    else:
        assert 0 <= result and result < low


@pytest.mark.parametrize(
    "seed",
    [-200, 400, 42, 0],
)
@pytest.mark.parametrize(
    ("low", "high"),
    [(1, None), (20, None), (200, 204), (-200, 10), (-40, -30), (0, 20), (-20, 0)],
)
@pytest.mark.parametrize(
    "size",
    [0, 10, (15, 20), (3, 4, 5), (6, 0, 7)],
)
def test_draw_random_integers(seed, low, high, size) -> None:
    from dioptra_builtins.random.sample import draw_random_integers

    rng = np.random.default_rng(seed if seed >= 0 else None)
    if seed < 0:
        seed = rng.bit_generator._seed_seq.entropy

    result: np.ndarray = draw_random_integers(rng, low, high, size)

    if size == 0:
        size = 1
    if type(size) is not tuple:
        size = (size,)
    assert result.shape == size



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform artifact endpoint operations."""
from __future__ import annotations

from typing import Any, Final, cast

import structlog
from flask_login import current_user
from injector import inject
from sqlalchemy import Integer, func, select
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.errors import (
    BackendDatabaseError,
    EntityDoesNotExistError,
    EntityExistsError,
    SortParameterValidationError,
)
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.groups.service import GroupIdService
from dioptra.restapi.v1.jobs.service import ExperimentJobIdService, JobIdService
from dioptra.restapi.v1.shared.search_parser import construct_sql_query_filters

LOGGER: BoundLogger = structlog.stdlib.get_logger()

RESOURCE_TYPE: Final[str] = "artifact"
SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "uri": lambda x: models.Artifact.uri.like(x, escape="/"),
    "description": lambda x: models.Artifact.description.like(x, escape="/"),
}
SORTABLE_FIELDS: Final[dict[str, Any]] = {
    "uri": models.Artifact.uri,
    "createdOn": models.Artifact.created_on,
    "lastModifiedOn": models.Resource.last_modified_on,
    "description": models.Artifact.description,
}


class ArtifactService(object):
    """The service methods for registering and managing artifacts by their unique id."""

    @inject
    def __init__(
        self,
        artifact_uri_service: ArtifactUriService,
        job_id_service: JobIdService,
        group_id_service: GroupIdService,
    ) -> None:
        """Initialize the artifact service.

        All arguments are provided via dependency injection.

        Args:
            artifact_uri_service: An ArtifactUriService object.
            job_id_service: A JobIdService object.
            group_id_service: A GroupIdService object.
        """
        self._artifact_uri_service = artifact_uri_service
        self._job_id_service = job_id_service
        self._group_id_service = group_id_service

    def create(
        self,
        uri: str,
        description: str,
        group_id: int,
        job_id: int,
        commit: bool = True,
        **kwargs,
    ) -> utils.ArtifactDict:
        """Create a new artifact.

        Args:
            uri: The uri of the artifact. This must be globally unique.
            description: The description of the artifact.
            group_id: The group that will own the artifact.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The newly created artifact object.

        Raises:
            EntityExistsError: If the artifact already exists.

        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        duplicate = self._artifact_uri_service.get(uri, log=log)
        if duplicate is not None:
            raise EntityExistsError(RESOURCE_TYPE, duplicate.resource_id, uri=uri)

        job_dict = cast(
            utils.JobDict,
            self._job_id_service.get(job_id, error_if_not_found=True, log=log),
        )
        job = job_dict["job"]

        group = self._group_id_service.get(group_id, error_if_not_found=True, log=log)

        resource = models.Resource(resource_type="artifact", owner=group)
        new_artifact = models.Artifact(
            uri=uri,
            description=description,
            resource=resource,
            creator=current_user,
        )
        job.children.append(new_artifact.resource)
        db.session.add(new_artifact)

        if commit:
            db.session.commit()
            log.debug(
                "Artifact registration successful", artifact_id=new_artifact.resource_id
            )

        return utils.ArtifactDict(artifact=new_artifact, has_draft=False)

    def get(
        self,
        group_id: int | None,
        search_string: str,
        page_index: int,
        page_length: int,
        sort_by_string: str,
        descending: bool,
        **kwargs,
    ) -> Any:
        """Fetch a list of artifacts, optionally filtering by search string and paging
        parameters.

        Args:
            group_id: A group ID used to filter results.
            search_string: A search string used to filter results.
            page_index: The index of the first group to be returned.
            page_length: The maximum number of artifacts to be returned.
            sort_by_string: The name of the column to sort.
            descending: Boolean indicating whether to sort by descending or not.

        Returns:
            A tuple containing a list of artifacts and the total number of artifacts
            matching the query.

        Raises:
            SearchNotImplementedError: If a search string is provided.
            BackendDatabaseError: If the database query returns a None when counting
                the number of artifacts.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get full list of artifacts")

        filters = list()

        if group_id is not None:
            filters.append(models.Resource.group_id == group_id)

        if search_string:
            filters.append(
                construct_sql_query_filters(search_string, SEARCHABLE_FIELDS)
            )

        stmt = (
            select(func.count(models.Artifact.resource_id))
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.Artifact.resource_snapshot_id,
            )
        )
        total_num_artifacts = db.session.scalars(stmt).first()

        if total_num_artifacts is None:
            log.error(
                "The database query returned a None when counting the number of "
                "groups when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_artifacts == 0:
            return [], total_num_artifacts

        # get latest artifact snapshots
        latest_artifacts_stmt = (
            select(models.Artifact)
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.Artifact.resource_snapshot_id,
            )
            .offset(page_index)
            .limit(page_length)
        )

        if sort_by_string and sort_by_string in SORTABLE_FIELDS:
            sort_column = SORTABLE_FIELDS[sort_by_string]
            if descending:
                sort_column = sort_column.desc()
            else:
                sort_column = sort_column.asc()
            latest_artifacts_stmt = latest_artifacts_stmt.order_by(sort_column)
        elif sort_by_string and sort_by_string not in SORTABLE_FIELDS:
            raise SortParameterValidationError(RESOURCE_TYPE, sort_by_string)

        artifacts = db.session.scalars(latest_artifacts_stmt).all()

        drafts_stmt = select(
            models.DraftResource.payload["resource_id"].as_string().cast(Integer)
        ).where(
            models.DraftResource.payload["resource_id"]
            .as_string()
            .cast(Integer)
            .in_(tuple(artifact.resource_id for artifact in artifacts)),
            models.DraftResource.user_id == current_user.user_id,
        )
        artifacts_dict: dict[int, utils.ArtifactDict] = {
            artifact.resource_id: utils.ArtifactDict(artifact=artifact, has_draft=False)
            for artifact in artifacts
        }
        for resource_id in db.session.scalars(drafts_stmt):
            artifacts_dict[resource_id]["has_draft"] = True

        return list(artifacts_dict.values()), total_num_artifacts


class JobArtifactService(object):
    """The service methods for registering and managing artifacts by their unique id."""

    @inject
    def __init__(
        self,
        artifact_service: ArtifactService,
        experiment_job_id_service: ExperimentJobIdService,
    ) -> None:
        """Initialize the artifact service.

        All arguments are provided via dependency injection.

        Args:
            artifact_service: An ArtifactService object.
            experiment_job_id_service: A JobIdService object.
        """
        self._artifact_service = artifact_service
        self._experiment_job_id_service = experiment_job_id_service

    def create(
        self,
        experiment_id: int,
        job_id: int,
        uri: str,
        description: str,
        commit: bool = True,
        **kwargs,
    ) -> utils.ArtifactDict:
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        job_dict = cast(
            utils.JobDict,
            self._experiment_job_id_service.get(
                experiment_id, job_id, error_if_not_found=True, log=log
            ),
        )
        job = job_dict["job"]

        resource = models.Resource(resource_type="artifact", owner=job.resource.owner)
        new_artifact = models.Artifact(
            uri=uri,
            description=description,
            resource=resource,
            creator=current_user,
        )
        job.children.append(new_artifact.resource)
        db.session.add(new_artifact)

        if commit:
            db.session.commit()
            log.debug(
                "Artifact registration successful", artifact_id=new_artifact.resource_id
            )

        return utils.ArtifactDict(artifact=new_artifact, has_draft=False)


class ArtifactUriService(object):
    """The service methods for managing artifacts by their uri."""

    def get(
        self,
        artifact_uri: str,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> models.Artifact | None:
        """Fetch an artifact by its unique uri.

        Args:
            artifact_uri: the unique uri of the artifact.
            error_if_not_found: If True, raise an error if the artifact is not found.
                Defaults to False.


        Returns:
            The artifact object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the artifact is not found and
                `error_if_not_found` is True.

        """

        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get artifact by id", artifact_uri=artifact_uri)

        stmt = (
            select(models.Artifact)
            .join(models.Resource)
            .where(
                models.Artifact.uri == artifact_uri,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.Artifact.resource_snapshot_id,
            )
        )
        artifact = db.session.scalars(stmt).first()

        if artifact is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(RESOURCE_TYPE, artifact_uri=artifact_uri)

            return None

        return artifact


class ArtifactIdService(object):
    """The service methods for retrieving artifacts by their unique id."""

    def get(
        self,
        artifact_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> utils.ArtifactDict | None:
        """Fetch a artifact by its unique id.

        Args:
            artifact_id: The unique id of the artifact.
            error_if_not_found: If True, raise an error if the artifact is not found.
                Defaults to False.

        Returns:
            The artifact object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the artifact is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get artifact by id", artifact_id=artifact_id)

        stmt = (
            select(models.Artifact)
            .join(models.Resource)
            .where(
                models.Artifact.resource_id == artifact_id,
                models.Artifact.resource_snapshot_id
                == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        artifact = db.session.scalars(stmt).first()

        if artifact is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(RESOURCE_TYPE, artifact_id=artifact_id)

            return None

        drafts_stmt = (
            select(models.DraftResource.draft_resource_id)
            .where(
                models.DraftResource.payload["resource_id"].as_string().cast(Integer)
                == artifact.resource_id,
                models.DraftResource.user_id == current_user.user_id,
            )
            .exists()
            .select()
        )
        has_draft = db.session.scalar(drafts_stmt)

        return utils.ArtifactDict(artifact=artifact, has_draft=has_draft)

    def modify(
        self,
        artifact_id: int,
        description: str,
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> utils.ArtifactDict | None:
        """Modify a artifact.

        Args:
            artifact_id: The unique id of the artifact.
            description: The new description of the artifact.
            error_if_not_found: If True, raise an error if the group is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated artifact object.

        Raises:
            EntityDoesNotExistError: If the artifact is not found and
                `error_if_not_found` is True.
            EntityExistsError: If the artifact name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        artifact_dict = self.get(
            artifact_id, error_if_not_found=error_if_not_found, log=log
        )

        if artifact_dict is None:
            return None

        artifact = artifact_dict["artifact"]
        has_draft = artifact_dict["has_draft"]

        new_artifact = models.Artifact(
            uri=artifact.uri,
            description=description,
            resource=artifact.resource,
            creator=current_user,
        )
        db.session.add(new_artifact)

        if commit:
            db.session.commit()
            log.debug(
                "Artifact modification successful",
                artifact_id=artifact_id,
                description=description,
            )

        return utils.ArtifactDict(artifact=new_artifact, has_draft=has_draft)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import List, Tuple

import numpy as np
import numpy.typing as npt
import structlog
from structlog.stdlib import BoundLogger

LOGGER: BoundLogger = structlog.stdlib.get_logger()


BoxesType = npt.NDArray
LabelsType = npt.NDArray


class AnnotationEncoding(metaclass=ABCMeta):
    @abstractmethod
    def decode(
        self, boxes: BoxesType, labels: LabelsType
    ) -> Tuple[List[List[float]], List[int]]:
        raise NotImplementedError

    @abstractmethod
    def encode(
        self, boxes: List[List[float]], labels: List[int]
    ) -> Tuple[BoxesType, LabelsType]:
        raise NotImplementedError


class NumpyAnnotationEncoding(AnnotationEncoding):
    def __init__(
        self, boxes_dtype: str = "float32", labels_dtype: str = "int32"
    ) -> None:
        self._boxes_dtype = boxes_dtype
        self._labels_dtype = labels_dtype

    def decode(
        self, boxes: npt.NDArray, labels: npt.NDArray
    ) -> Tuple[List[List[float]], List[int]]:
        decoded_boxes: List[List[float]] = [
            [float(coord) for coord in x] for x in boxes.tolist()
        ]
        decoded_labels: List[int] = [int(x) for x in labels.tolist()]

        return decoded_boxes, decoded_labels

    def encode(
        self, boxes: List[List[float]], labels: List[int]
    ) -> Tuple[npt.NDArray, npt.NDArray]:
        return np.array(boxes, dtype=self._boxes_dtype), np.array(
            labels, dtype=self._labels_dtype
        )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from typing import Final

from sqlalchemy import insert
from sqlalchemy.orm import Session

from dioptra.restapi.db import models

ENTRY_POINT_PARAMETER_TYPES: Final[list[dict[str, str]]] = [
    {"parameter_type": "string"},
    {"parameter_type": "float"},
    {"parameter_type": "integer"},
    {"parameter_type": "boolean"},
    {"parameter_type": "list"},
    {"parameter_type": "mapping"},
]
JOB_STATUS_TYPES: Final[list[dict[str, str]]] = [
    {"status": "queued"},
    {"status": "started"},
    {"status": "deferred"},
    {"status": "finished"},
    {"status": "failed"},
]
USER_LOCK_TYPES: Final[list[dict[str, str]]] = [
    {"user_lock_type": "delete"},
]
GROUP_LOCK_TYPES: Final[list[dict[str, str]]] = [
    {"group_lock_type": "delete"},
]
RESOURCE_LOCK_TYPES: Final[list[dict[str, str]]] = [
    {"resource_lock_type": "delete"},
    {"resource_lock_type": "readonly"},
]
RESOURCE_TYPES: Final[list[dict[str, str]]] = [
    {"resource_type": "artifact"},
    {"resource_type": "entry_point"},
    {"resource_type": "experiment"},
    {"resource_type": "job"},
    {"resource_type": "ml_model"},
    {"resource_type": "plugin"},
    {"resource_type": "plugin_file"},
    {"resource_type": "plugin_task_parameter_type"},
    {"resource_type": "queue"},
    {"resource_type": "resource_snapshot"},
    {"resource_type": "ml_model_version"},
]
RESOURCE_DEPENDENCY_TYPES: Final[list[dict[str, str]]] = [
    {"parent_resource_type": "experiment", "child_resource_type": "entry_point"},
    {"parent_resource_type": "entry_point", "child_resource_type": "plugin"},
    {"parent_resource_type": "entry_point", "child_resource_type": "queue"},
    {"parent_resource_type": "plugin", "child_resource_type": "plugin_file"},
    {"parent_resource_type": "job", "child_resource_type": "artifact"},
    {"parent_resource_type": "job", "child_resource_type": "job"},
    {"parent_resource_type": "ml_model", "child_resource_type": "ml_model_version"},
]


def setup_ontology(session: Session) -> None:
    """Setup the ontology tables in the database.

    Args:
        session: The SQLAlchemy session object.
    """
    stmts = [
        insert(models.entry_point_parameter_types_table).values(
            ENTRY_POINT_PARAMETER_TYPES
        ),
        insert(models.job_status_types_table).values(JOB_STATUS_TYPES),
        insert(models.user_lock_types_table).values(USER_LOCK_TYPES),
        insert(models.group_lock_types_table).values(GROUP_LOCK_TYPES),
        insert(models.resource_lock_types_table).values(RESOURCE_LOCK_TYPES),
        insert(models.resource_types_table).values(RESOURCE_TYPES),
        insert(models.resource_dependency_types_table).values(
            RESOURCE_DEPENDENCY_TYPES
        ),
    ]

    for stmt in stmts:
        session.execute(stmt)

    session.commit()



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from typing import Final

import structlog
from redis import Redis
from rq.queue import Queue as RQQueue
from structlog.stdlib import BoundLogger

LOGGER: BoundLogger = structlog.stdlib.get_logger()

TIMEOUT_24_HOURS: Final[int] = 24 * 3600
RUN_V1_DIOPTRA_JOB_FUNC: Final[str] = "dioptra.rq.tasks.run_v1_dioptra_job"


class RQServiceV1(object):
    def __init__(self, redis: Redis) -> None:
        self._redis = redis

    def submit(
        self,
        job_id: int,
        experiment_id: int,
        queue: str,
        timeout: str | None = None,
    ):
        log: BoundLogger = LOGGER.new()

        cmd_kwargs = {
            "job_id": job_id,
            "experiment_id": experiment_id,
        }

        log.info(
            "Enqueuing job",
            function=RUN_V1_DIOPTRA_JOB_FUNC,
            job_id=job_id,
            cmd_kwargs=cmd_kwargs,
            timeout=timeout,
        )

        q = RQQueue(queue, default_timeout=TIMEOUT_24_HOURS, connection=self._redis)
        q.enqueue(
            RUN_V1_DIOPTRA_JOB_FUNC,
            kwargs=cmd_kwargs,
            job_id=str(job_id),
            timeout=timeout,
        )



# Code below is copied from Flask-Injector library.  It has the following
# license:
#
# Copyright (c) 2012, Alec Thomas
# Copyright (c) 2015 Smarkets Limited <support@smarkets.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#  - Neither the name of SwapOff.org nor the names of its contributors may
#    be used to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Tell flake8 to ignore this file.  It is mostly just copy-pastes from
# Flask-Injector.
# flake8: noqa

from typing import Any, Dict

import flask
from injector import Injector, Provider, Scope, ScopeDecorator
from werkzeug.local import Local, LocalManager


class CachedProviderWrapper(Provider):
    def __init__(self, old_provider: Provider) -> None:
        self._old_provider = old_provider
        self._cache = {}  # type: Dict[int, Any]

    def get(self, injector: Injector) -> Any:
        key = id(injector)
        try:
            return self._cache[key]
        except KeyError:
            instance = self._cache[key] = self._old_provider.get(injector)
            return instance


class RequestScope(Scope):
    """A scope whose object lifetime is tied to a request.

    @request
    class Session:
        pass
    """

    # We don't want to assign here, just provide type hints
    if False:
        _local_manager = None  # type: LocalManager
        _locals = None  # type: Any

    def cleanup(self) -> None:
        self._local_manager.cleanup()

    def prepare(self) -> None:
        self._locals.scope = {}

    def configure(self) -> None:
        self._locals = Local()
        self._local_manager = LocalManager([self._locals])
        self.prepare()

    def get(self, key: Any, provider: Provider) -> Any:
        try:
            return self._locals.scope[key]
        except KeyError:
            new_provider = self._locals.scope[key] = CachedProviderWrapper(provider)
            return new_provider


request = ScopeDecorator(RequestScope)


def set_request_scope_callbacks(app: flask.Flask, injector: Injector) -> None:
    """
    Set callbacks to enable request scoping behavior: initialize at the
    beginning of request handling, and cleanup at the end.

    Args:
        app: A Flask app
        injector: An injector, used to get the RequestScope object
    """

    def reset_request_scope_before(*args: Any, **kwargs: Any) -> None:
        injector.get(RequestScope).prepare()

    def global_reset_request_scope_after(*args: Any, **kwargs: Any) -> None:
        blueprint = flask.request.blueprint
        # If current blueprint has teardown_request_funcs associated with it we know there may be
        # a some teardown request handlers we need to inject into, so we can't reset the scope just yet.
        # We'll leave it to blueprint_reset_request_scope_after to do the job which we know will run
        # later and we know it'll run after any teardown_request handlers we may want to inject into.
        if blueprint is None or blueprint not in app.teardown_request_funcs:
            injector.get(RequestScope).cleanup()

    def blueprint_reset_request_scope_after(*args: Any, **kwargs: Any) -> None:
        # If we got here we truly know this is the last teardown handler, which means we can reset the
        # scope unconditionally.
        injector.get(RequestScope).cleanup()

    app.before_request_funcs.setdefault(None, []).insert(0, reset_request_scope_before)
    # We're accessing Flask internals here as the app.teardown_request decorator appends to a list of
    # handlers but Flask itself reverses the list when it executes them. To allow injecting request-scoped
    # dependencies into teardown_request handlers we need to run our teardown_request handler after them.
    # Also see https://github.com/alecthomas/flask_injector/issues/42 where it was reported.
    # Secondly, we need to handle blueprints. Flask first executes non-blueprint teardown handlers in
    # reverse order and only then executes blueprint-associated teardown handlers in reverse order,
    # which means we can't just set on non-blueprint teardown handler, but we need to set both.
    # In non-blueprint teardown handler we check if a blueprint handler will run – if so, we do nothing
    # there and leave it to the blueprint teardown handler.
    #
    # We need the None key to be present in the dictionary so that the dictionary iteration always yields
    # None as well. We *always* have to set the global teardown request.
    app.teardown_request_funcs.setdefault(None, []).insert(
        0, global_reset_request_scope_after
    )
    for bp, functions in app.teardown_request_funcs.items():
        if bp is not None:
            functions.insert(0, blueprint_reset_request_scope_after)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import argparse
import sys

import yaml

import dioptra.task_engine.validation


def parse_args() -> argparse.Namespace:
    arg_parser = argparse.ArgumentParser(
        description="""
        Simple commandline tool to statically validate a declarative
        experiment description file.
        """,
        epilog="""
        Exit status is 0 if no issues were found; 1 if there were issues; 2
        if an exception occurs.
        """,
    )

    arg_parser.add_argument(
        "file",
        help="""The file to validate""",
        type=argparse.FileType("r", encoding="utf-8"),
    )

    arg_parser.add_argument(
        "-q",
        "--quiet",
        help="""Be quiet.  If given twice, silence stack traces too.""",
        action="count",
        default=0,
    )

    return arg_parser.parse_args()


def main() -> int:
    exit_status = 0
    args = parse_args()

    try:
        experiment_desc = yaml.safe_load(args.file)
        issues = dioptra.task_engine.validation.validate(experiment_desc)
        if issues:
            exit_status = 1
            if args.quiet < 1:
                print("Issues:")
                print()
                for i, issue in enumerate(issues):
                    print(i + 1, ". ", issue, sep="")
                    print()

        else:
            if args.quiet < 1:
                print("No issues!")

    except Exception:
        exit_status = 2
        if args.quiet < 2:
            sys.excepthook(*sys.exc_info())

    return exit_status


if __name__ == "__main__":
    sys.exit(main())



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from pathlib import Path
from typing import Dict, Iterable, List, Tuple, Union, cast
from xml.etree import ElementTree

import structlog
from structlog.stdlib import BoundLogger

from .annotation_data import AnnotationData
from .encodings import AnnotationEncoding, BoxesType, LabelsType

LOGGER: BoundLogger = structlog.stdlib.get_logger()


class PascalVOCAnnotationData(AnnotationData):
    def __init__(
        self,
        labels: Iterable[str],
        encoding: AnnotationEncoding,
    ) -> None:
        self._labels = {x: idx for idx, x in enumerate(labels)}
        self._encoding = encoding

    @property
    def labels(self) -> Dict[str, int]:
        return self._labels

    def get(self, y: Union[Path, bytes, str]) -> Tuple[BoxesType, LabelsType]:
        boxes, classes = self.read_file(filepath=y)
        encoded_boxes, encoded_classes = self._encoding.encode(boxes, classes)

        return encoded_boxes, encoded_classes

    def read_file(
        self, filepath: Union[Path, bytes, str]
    ) -> Tuple[list[list[float]], list[int]]:
        # Load and parse the file
        filepath = filepath.decode() if isinstance(filepath, bytes) else str(filepath)
        tree: ElementTree.ElementTree = ElementTree.parse(filepath)

        # Get the root of the document
        root = tree.getroot()
        boxes: List[List[float]] = list()
        classes: List[int] = list()

        # Get width and height of an image
        width_element: ElementTree.Element | None = root.find(".//size/width")
        height_element: ElementTree.Element | None = root.find(".//size/height")
        width: int = (
            int(cast(str, width_element.text)) if width_element is not None else -1
        )
        height: int = (
            int(cast(str, height_element.text)) if height_element is not None else -1
        )

        # Extract each bounding box
        for box in root.findall(".//object"):
            class_name_element: ElementTree.Element | None = box.find("name")
            class_name: str = (
                cast(str, class_name_element.text)
                if class_name_element is not None
                else ""
            )
            class_id = self._labels[class_name]
            xmin_element: ElementTree.Element | None = box.find("bndbox/xmin")
            ymin_element: ElementTree.Element | None = box.find("bndbox/ymin")
            xmax_element: ElementTree.Element | None = box.find("bndbox/xmax")
            ymax_element: ElementTree.Element | None = box.find("bndbox/ymax")
            xmin = int(cast(str, xmin_element.text)) if xmin_element is not None else -1
            ymin = int(cast(str, ymin_element.text)) if ymin_element is not None else -1
            xmax = int(cast(str, xmax_element.text)) if xmax_element is not None else -1
            ymax = int(cast(str, ymax_element.text)) if ymax_element is not None else -1
            coordinates = [xmin / width, ymin / height, xmax / width, ymax / height]
            boxes.append(coordinates)
            classes.append(class_id)

        return boxes, classes



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Shared actions for REST API unit tests.

This module contains shared actions used across test suites for each of the REST
API endpoints.
"""
from typing import Any

from flask.testing import FlaskClient
from werkzeug.test import TestResponse

from dioptra.restapi.routes import (
    V1_ARTIFACTS_ROUTE,
    V1_AUTH_ROUTE,
    V1_ENTRYPOINTS_ROUTE,
    V1_EXPERIMENTS_ROUTE,
    V1_GROUPS_ROUTE,
    V1_JOBS_ROUTE,
    V1_MODELS_ROUTE,
    V1_PLUGIN_PARAMETER_TYPES_ROUTE,
    V1_PLUGINS_ROUTE,
    V1_QUEUES_ROUTE,
    V1_ROOT,
    V1_TAGS_ROUTE,
    V1_USERS_ROUTE,
)


def login(client: FlaskClient, username: str, password: str) -> TestResponse:
    """Login a user using the API.

    Args:
        client: The Flask test client.
        username: The username of the user to be logged in.
        password: The password of the user to be logged in.

    Returns:
        The response from the API.
    """
    return client.post(
        f"/{V1_ROOT}/{V1_AUTH_ROUTE}/login",
        json={"username": username, "password": password},
    )


def register_entrypoint(
    client: FlaskClient,
    name: str,
    description: str,
    group_id: int,
    task_graph: str,
    parameters: list[dict[str, Any]],
    plugin_ids: list[int],
    queue_ids: list[int],
) -> TestResponse:
    """Register an entrypoint using the API.

    Args:
        client: The Flask test client.
        name: The name to assign to the new entrypoint.
        description: The description for the new entrypoint.
        group_id: The group to create the new entrypoint in.
        task_graph: A graph of the tasks to perform in a entry point.
        parameters: A list of entrypoint parameters.

    Returns:
        The response from the API.
    """
    payload: dict[str, Any] = {
        "name": name,
        "group": group_id,
        "taskGraph": task_graph,
        "parameters": parameters,
        "plugins": plugin_ids,
        "queues": queue_ids,
    }

    if description:
        payload["description"] = description

    return client.post(
        f"/{V1_ROOT}/{V1_ENTRYPOINTS_ROUTE}/",
        json=payload,
        follow_redirects=True,
    )


def register_experiment(
    client: FlaskClient,
    name: str,
    group_id: int,
    entrypoint_ids: list[int] | None = None,
    description: str | None = None,
) -> TestResponse:
    """Register an experiment using the API.

    Args:
        client: The Flask test client.
        name: The name to assign to the new experiment.
        group_id: The group to create the new experiment in.
        entrypoints: The entrypoint IDs for the new experiment.
        description: The description of the new experiment.

    Returns:
        The response from the API.
    """

    payload = {"name": name, "group": group_id}

    if description is not None:
        payload["description"] = description
    if entrypoint_ids is not None:
        payload["entrypoints"] = entrypoint_ids

    return client.post(
        f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}/",
        json=payload,
        follow_redirects=True,
    )


def register_job(
    client: FlaskClient,
    queue_id: int,
    experiment_id: int,
    entrypoint_id: int,
    values: dict[str, str],
    description: str | None = None,
    timeout: str | None = None,
) -> TestResponse:
    """Register a job using the API.

    Args:
        client:
        queue_id:
        experiment_id:
        entrypoint_id:
        description:
        values:
        timeout:

    Returns:
        The response from the API.
    """
    payload: dict[str, Any] = {
        "queue": queue_id,
        "entrypoint": entrypoint_id,
        "values": values,
    }

    if description:
        payload["description"] = description

    if timeout:
        payload["timeout"] = timeout

    return client.post(
        f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}/{experiment_id}/{V1_JOBS_ROUTE}",
        json=payload,
        follow_redirects=True,
    )


def register_user(
    client: FlaskClient, username: str, email: str, password: str
) -> TestResponse:
    """Register a user using the API.

    Args:
        client: The Flask test client.
        username: The username to assign to the new user.
        email: The email to assign to the new user.
        password: The password to set for the new user.

    Returns:
        The response from the API.
    """
    return client.post(
        f"/{V1_ROOT}/{V1_USERS_ROUTE}",
        json={
            "username": username,
            "email": email,
            "password": password,
            "confirmPassword": password,
        },
        follow_redirects=True,
    )


def register_tag(
    client: FlaskClient,
    name: str,
    group_id: int,
) -> TestResponse:
    """Register a tag using the API.
    Args:
        client: The Flask test client.
        name: The name to assign to the new tag.
        group_id: The group to create the new tag in.
    Returns:
        The response from the API.
    """

    payload = {"name": name, "group": group_id}

    return client.post(
        f"/{V1_ROOT}/{V1_TAGS_ROUTE}/",
        json=payload,
        follow_redirects=True,
    )


def register_artifact(
    client: FlaskClient,
    uri: str,
    job_id: int,
    group_id: int,
    description: str | None = None,
) -> TestResponse:
    """Register an artifact using the API.

    Args:
        client: The Flask test client.
        uri: The URI of the artifact
        job_id: The job to create the new artifact.
        group_id: The group to create the new artifact in.
        description: The description of the new artifact.

    Returns:
        The response from the API.
    """
    payload = {"uri": uri, "job": job_id, "group": group_id}

    if description is not None:
        payload["description"] = description

    return client.post(
        f"/{V1_ROOT}/{V1_ARTIFACTS_ROUTE}/",
        json=payload,
        follow_redirects=True,
    )


def register_model(
    client: FlaskClient,
    name: str,
    group_id: int,
    description: str | None = None,
) -> TestResponse:
    """Register a model using the API.

    Args:
        client: The Flask test client.
        name: The name to assign to the new model.
        group_id: The group to create the new model in.
        description: The description of the new model.

    Returns:
        The response from the API.
    """
    payload = {"name": name, "group": group_id, "description": description}

    return client.post(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}/",
        json=payload,
        follow_redirects=True,
    )


def register_model_version(
    client: FlaskClient,
    model_id: int,
    artifact_id: int,
    description: str | None = None,
) -> TestResponse:
    """Register a model version using the API.

    Args:
        client: The Flask test client.
        model_id: The id of the model to create a new version of.
        artifact_id: The id of artifact representing the new model version.
        description: The description of the new model version.

    Returns:
        The response from the API.
    """
    payload = {"artifact": artifact_id, "description": description}

    return client.post(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}/{model_id}/versions",
        json=payload,
        follow_redirects=True,
    )


def register_plugin(
    client: FlaskClient,
    name: str,
    description: str,
    group_id: int,
) -> TestResponse:
    """Register a plugin using the API.

    Args:
        client: The Flask test client.
        name: The name to assign to the new plugin.
        description: The description of the new plugin.
        group_id: The group to create the new plugin in.

    Returns:
        The response from the API.
    """
    payload = {"name": name, "description": description, "group": group_id}

    return client.post(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/",
        json=payload,
        follow_redirects=True,
    )


def register_plugin_file(
    client: FlaskClient,
    plugin_id: int,
    description: str,
    filename: str,
    contents: str,
    tasks: list[dict[str, Any]] | None = None,
) -> TestResponse:
    """Register a plugin file using the API.

    Args:
        client: The Flask test client.
        plugin_id: The plugin resource to create the file in.
        filename: The name of the plugin file.
        contents: The contents of the file containing imports, functions,
            structures, etc.
        tasks: Tasks associated with the plugin file resource.
        description: The description of the plugin file.

    Returns:
        The response from the API.
    """
    payload: dict[str, Any] = {
        "filename": filename,
        "contents": contents,
        "description": description,
        "tasks": tasks or [],
    }

    return client.post(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}/files",
        json=payload,
        follow_redirects=True,
    )


def register_plugin_parameter_type(
    client: FlaskClient,
    name: str,
    group_id: int,
    structure: dict[str, Any] | None = None,
    description: str | None = None,
) -> TestResponse:
    """Create a Plugin Parameter Type using the API.

    Args:
        client: The Flask test client.
        name: The name of the plugin parameter type to be created.
        group_id: The group to create the new plugin parameter type in.
        structure: Optional JSON-type field for further constraining a type's structure.

    Returns:
        The response from the API.
    """
    payload = {"name": name, "group": group_id}

    if structure:
        payload["structure"] = structure

    if description:
        payload["description"] = description

    return client.post(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}",
        json=payload,
        follow_redirects=True,
    )


def register_queue(
    client: FlaskClient,
    name: str,
    description: str,
    group_id: int,
) -> TestResponse:
    """Register a queue using the API.

    Args:
        client: The Flask test client.
        name: The name to assign to the new queue.
        group_id: The group to create the new queue in.
        description: The description of the new queue.

    Returns:
        The response from the API.
    """
    payload = {"name": name, "description": description, "group": group_id}

    return client.post(
        f"/{V1_ROOT}/{V1_QUEUES_ROUTE}/",
        json=payload,
        follow_redirects=True,
    )


def register_group(
    client: FlaskClient,
    name: str,
) -> TestResponse:
    """Register a group using the API.

    Args:
        client: The Flask test client.
        name: The name to assign to the new group.

    Returns:
        The response from the API.
    """
    payload: dict[str, Any] = {"name": name}

    return client.post(
        f"/{V1_ROOT}/{V1_GROUPS_ROUTE}/",
        json=payload,
        follow_redirects=True,
    )


def get_public_group(client: FlaskClient) -> TestResponse:
    """Get the public group.

    Args:
        client: The Flask test client.

    Returns:
        The response from the API.
    """
    return client.get(f"/{V1_ROOT}/{V1_GROUPS_ROUTE}/1", follow_redirects=True)


def get_model(client: FlaskClient, model_id: int) -> TestResponse:
    response = client.get(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}/{model_id}",
        follow_redirects=True,
    )
    return response


def get_plugin_parameter_types(client: FlaskClient) -> TestResponse:
    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}",
        follow_redirects=True,
    )
    return response


def get_draft(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
) -> TestResponse:
    """Get the draft of the resource with the provided unique ID.

    Args:
        client: The Flask test client.
        resource_route: The route of the resource type for the draft.
        resource_id: The id of the queue to rename.

    Returns:
        The response from the API.
    """

    return client.get(f"/{V1_ROOT}/{resource_route}/{resource_id}/draft")


def create_existing_resource_draft(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
    payload: dict[str, Any],
) -> TestResponse:
    """Create a draft of the resource with the provided unique ID.

    Args:
        client: The Flask test client.
        resource_route: The route of the resource type for the draft.
        resource_id: The id of the resource to create draft of.
        payload: The contents of the draft resource.

    Returns:
        The response from the API.
    """
    return client.post(
        f"/{V1_ROOT}/{resource_route}/{resource_id}/draft",
        json=payload,
        follow_redirects=True,
    )


def modify_existing_resource_draft(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
    payload: dict[str, Any],
) -> TestResponse:
    """Modify the draft of the resource with the provided unique ID.

    Args:
        client: The Flask test client.
        resource_route: The route of the resource type for the draft.
        resource_id: The id of the resource to modify.
        payload: The new contents of the draft resource.

    Returns:
        The response from the API.
    """
    return client.put(
        f"/{V1_ROOT}/{resource_route}/{resource_id}/draft",
        json=payload,
        follow_redirects=True,
    )


def delete_existing_resource_draft(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
) -> TestResponse:
    """Delete the draft of the resource with the provided unique ID.

    Args:
        client: The Flask test client.
        resource_route: The route of the resource type for the draft.
        resource_id: The id of the resource draft to delete.

    Returns:
        The response from the API.
    """

    return client.delete(
        f"/{V1_ROOT}/{resource_route}/{resource_id}/draft",
        follow_redirects=True,
    )


def get_drafts(
    client: FlaskClient,
    resource_route: str,
) -> TestResponse:
    """Get a list of drafts for the resource type

    Args:
        client: The Flask test client.
        resource_route: The route of the resource type for the draft.

    Returns:
        The response from the API.
    """

    return client.get(f"/{V1_ROOT}/{resource_route}/drafts")


def create_new_resource_draft(
    client: FlaskClient,
    resource_route: str,
    group_id: int | None,
    payload: dict[str, Any],
) -> TestResponse:
    """Create a draft resource using the API.

    Args:
        client: The Flask test client.
        resource_route: The route of the resource type for the draft.
        payload: The contents of the draft resource.

    Returns:
        The response from the API.
    """
    if group_id is not None:
        payload = {**payload, "group": group_id}

    return client.post(
        f"/{V1_ROOT}/{resource_route}/drafts",
        json=payload,
        follow_redirects=True,
    )


def modify_new_resource_draft(
    client: FlaskClient,
    resource_route: str,
    draft_id: int,
    payload: dict[str, Any],
) -> TestResponse:
    """Modify a draft resource using the API.

    Args:
        client: The Flask test client.
        draft_id: The id of the draft to modify.
        payload: The new contents of the draft resource.

    Returns:
        The response from the API.
    """
    return client.put(
        f"/{V1_ROOT}/{resource_route}/drafts/{draft_id}",
        json=payload,
        follow_redirects=True,
    )


def delete_new_resource_draft(
    client: FlaskClient,
    resource_route: str,
    draft_id: int,
) -> TestResponse:
    """Delete a draft resource using the API.

    Args:
        client: The Flask test client.
        resource_id: The id of the resource to modify.
        draft_id: The id of the draft to modify.

    Returns:
        The response from the API.
    """
    return client.delete(
        f"/{V1_ROOT}/{resource_route}/drafts/{draft_id}",
        follow_redirects=True,
    )


def get_tags(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
) -> TestResponse:
    """Get a list of tags for the resource type

    Args:
        client: The Flask test client.
        resource_route: The route of the resource type.
        resource_id: The id of the resource to append tags to.

    Returns:
        The response from the API.
    """

    return client.get(
        f"/{V1_ROOT}/{resource_route}/{resource_id}/tags", follow_redirects=True
    )


def append_tags(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
    tag_ids: list[int],
) -> TestResponse:
    """Append tags to the resource with the provided unique ID.

    Args:
        client: The Flask test client.
        resource_route: The route of the resource type.
        resource_id: The id of the resource to append tags to.
        payload: The contents of the tags resource.

    Returns:
        The response from the API.
    """
    return client.post(
        f"/{V1_ROOT}/{resource_route}/{resource_id}/tags",
        json={"ids": tag_ids},
        follow_redirects=True,
    )


def modify_tags(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
    tag_ids: list[int],
) -> TestResponse:
    """Modify the tags of the resource with the provided unique ID.

    Args:
        client: The Flask test client.
        resource_route: The route of the resource type.
        resource_id: The id of the resource to modify.
        tag_ids: The new tag_ids for the resource

    Returns:
        The response from the API.
    """
    return client.put(
        f"/{V1_ROOT}/{resource_route}/{resource_id}/tags",
        json={"ids": tag_ids},
        follow_redirects=True,
    )


def remove_tags(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
) -> TestResponse:
    """Remove the tags of the resource with the provided unique ID.

    Args:
        client: The Flask test client.
        resource_route: The route of the resource type.
        resource_id: The id of the resource to remove tags from.

    Returns:
        The response from the API.
    """

    return client.delete(
        f"/{V1_ROOT}/{resource_route}/{resource_id}/tags",
        follow_redirects=True,
    )


def remove_tag(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
    tag_id: int,
) -> TestResponse:
    """Remove tag from the resource with the provided unique ID.

    Args:
        client: The Flask test client.
        resource_route: The route of the resource type.
        resource_id: The id of the resource to remove a tag from.
        tag_id: The id of the tag to delete.

    Returns:
        The response from the API.
    """

    return client.delete(
        f"/{V1_ROOT}/{resource_route}/{resource_id}/tags/{tag_id}",
        follow_redirects=True,
    )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Test suite for group operations.
This module contains a set of tests that validate the CRUD operations and additional
functionalities for the group entity. The tests ensure that the groups can be
registered, queried, and renamed as expected through the REST API.
"""
from typing import Any

import pytest
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy
from werkzeug.test import TestResponse

from dioptra.restapi.routes import V1_GROUPS_ROUTE, V1_ROOT

from ..lib import actions, helpers

# -- Actions ---------------------------------------------------------------------------


def modify_group(
    client: FlaskClient,
    group_id: int,
    new_name: str,
) -> TestResponse:
    """Rename a group using the API.

    Args:
        client: The Flask test client.
        group_id: The id of the group to rename.
        new_name: The new name to assign to the group.

    Returns:
        The response from the API.
    """
    payload: dict[str, Any] = {"name": new_name}

    return client.put(
        f"/{V1_ROOT}/{V1_GROUPS_ROUTE}/{group_id}",
        json=payload,
        follow_redirects=True,
    )


# -- Assertions ------------------------------------------------------------------------


def assert_group_response_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    """Assert that group response contents is valid.

    Args:
        response: The actual response from the API.
        expected_contents: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response or if the response contents is not
            valid.
    """
    expected_keys = {
        "id",
        "name",
        "user",
        "members",
        "createdOn",
        "lastModifiedOn",
    }
    assert set(response.keys()) == expected_keys

    # Validate the non-Ref fields
    assert isinstance(response["id"], int)
    assert isinstance(response["name"], str)
    assert isinstance(response["createdOn"], str)
    assert isinstance(response["lastModifiedOn"], str)

    assert response["name"] == expected_contents["name"]

    assert helpers.is_iso_format(response["createdOn"])
    assert helpers.is_iso_format(response["lastModifiedOn"])

    # Validate the UserRef structure
    assert isinstance(response["user"]["id"], int)
    assert isinstance(response["user"]["username"], str)
    assert isinstance(response["user"]["url"], str)
    assert response["user"]["id"] == expected_contents["user_id"]

    # Validate the that each member is a GroupMember
    for member in response["members"]:

        # Validate the UserRef structure for member
        assert isinstance(member["user"]["id"], int)
        assert isinstance(member["user"]["username"], str)
        assert isinstance(member["user"]["url"], str)

        # Validate the GroupRef structure for member
        assert isinstance(member["group"]["id"], int)
        assert isinstance(member["group"]["name"], str)

        # Validate permissions
        assert isinstance(member["permissions"]["owner"], bool)
        assert isinstance(member["permissions"]["admin"], bool)
        assert isinstance(member["permissions"]["read"], bool)
        assert isinstance(member["permissions"]["write"], bool)
        assert isinstance(member["permissions"]["shareRead"], bool)
        assert isinstance(member["permissions"]["shareWrite"], bool)


def assert_retrieving_group_by_id_works(
    client: FlaskClient,
    group_id: int,
    expected: dict[str, Any],
) -> None:
    """Assert that retrieving a group by id works.

    Args:
        client: The Flask test client.
        group_id: The id of the group to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_GROUPS_ROUTE}/{group_id}", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_retrieving_groups_works(
    client: FlaskClient,
    expected: list[dict[str, Any]],
    search: str | None = None,
    paging_info: dict[str, Any] | None = None,
) -> None:
    """Assert that retrieving all groups works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.
        search: The search string used in query parameters.
        paging_info: The paging information used in query parameters.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    if search is not None:
        query_string["search"] = search

    if paging_info is not None:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{V1_GROUPS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_registering_existing_group_name_fails(
    client: FlaskClient, name: str
) -> None:
    """Assert that registering a group with an existing name fails.

    Args:
        client: The Flask test client.
        name: The name to assign to the new group.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = actions.register_group(client, name=name)
    assert response.status_code == 400


def assert_group_name_matches_expected_name(
    client: FlaskClient, group_id: int, expected_name: str
) -> None:
    """Assert that the name of a group matches the expected name.

    Args:
        client: The Flask test client.
        group_id: The id of the group to retrieve.
        expected_name: The expected name of the group.

    Raises:
        AssertionError: If the response status code is not 200 or if the name of the
            group does not match the expected name.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_GROUPS_ROUTE}/{group_id}",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["name"] == expected_name


def assert_cannot_rename_group_with_existing_name(
    client: FlaskClient,
    group_id: int,
    existing_name: str,
) -> None:
    """Assert that renaming a group with an existing name fails.

    Args:
        client: The Flask test client.
        group_id: The id of the group to rename.
        name: The name of an existing group.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = modify_group(
        client=client,
        group_id=group_id,
        new_name=existing_name,
    )
    assert response.status_code == 400


# -- Tests -----------------------------------------------------------------------------


@pytest.mark.v1_test
def test_create_group(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> None:
    """Test that groups can be correctly registered and retrieved using the API.

    Given an authenticated user, this test validates the following sequence of actions:

    - The user registers a group named "new_group".
    - The response is valid matches the expected values given the registration request.
    - The user is able to retrieve information about the group using the group id.

    Notes:
        The group id is generated in the backend during the POST and is returned in the
        response.
    """
    name = "new_group"
    user_id = auth_account["id"]
    group_expected = actions.register_group(client, name=name).get_json()
    assert_group_response_contents_matches_expectations(
        response=group_expected,
        expected_contents={
            "name": name,
            "user_id": user_id,
        },
    )
    assert_retrieving_group_by_id_works(
        client, group_id=group_expected["id"], expected=group_expected
    )


def test_group_get_all(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_groups: dict[str, Any],
) -> None:
    """Test that all users can be retrieved.

    Given an authenticated user and registered groups, this test validates the following
    sequence of actions:

    - The user is able to retrieve a list of all registered groups.
    - The returned list of groups matches the full list of registered groups.
    """
    group_expected_list = list(registered_groups.values())
    assert_retrieving_groups_works(client, expected=group_expected_list)


def test_group_search_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_groups: dict[str, Any],
) -> None:
    """Test that groups can be queried with a search term.

    Given an authenticated user and registered groups, this test validates the following
    sequence of actions:

    - The user is able to retrieve a list of all registered groups with a name
      that match the wildcard '*'.
    - The returned list of groups matches the expected matches from the query.
    """
    group_expected_list = [registered_groups["public"]]
    assert_retrieving_groups_works(
        client, expected=group_expected_list, search="name:public"
    )
    assert_retrieving_groups_works(
        client, expected=group_expected_list, search="public"
    )
    assert_retrieving_groups_works(client, expected=group_expected_list, search="pub*")
    assert_retrieving_groups_works(client, expected=[], search="name:pub")


@pytest.mark.v1_test
def test_cannot_register_existing_group_name(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_groups: dict[str, Any],
) -> None:
    """Test that registering a group with an existing name fails.

    Given an authenticated user and registered groups, this test validates the following
    sequence of actions:

    - The user attempts to register a second group with the same name.
    - The request fails with an appropriate error message and response code.
    """
    existing_group = registered_groups["group1"]
    assert_registering_existing_group_name_fails(client, name=existing_group["name"])


@pytest.mark.v1_test
def test_rename_group(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_groups: dict[str, Any],
) -> None:
    """Test that a group can be renamed.

    Given an authenticated user and registered groups, this test validates the following
    sequence of actions:

    - The user issues a request to change the name of a group.
    - The user retrieves information about the same group and it reflects the name
      change.
    - The user issues a request to change the name of a group to an existing group's
      name.
    - The request fails with an appropriate error message and response code.
    """
    updated_group_name = "updated_name"
    group_to_rename = registered_groups["group1"]
    existing_group = registered_groups["group2"]

    modify_group(client, group_id=group_to_rename["id"], new_name=updated_group_name)
    assert_group_name_matches_expected_name(
        client, group_id=group_to_rename["id"], expected_name=updated_group_name
    )
    assert_cannot_rename_group_with_existing_name(
        client,
        group_id=group_to_rename["id"],
        existing_name=existing_group["name"],
    )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform tag endpoint operations."""
from __future__ import annotations

from typing import Any, cast

import structlog
from injector import ClassAssistedBuilder, inject
from sqlalchemy import select
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.errors import EntityDoesNotExistError
from dioptra.restapi.v1.tags.service import RESOURCE_TYPE as TAG_RESOURCE_TYPE
from dioptra.restapi.v1.tags.service import TagIdService

LOGGER: BoundLogger = structlog.stdlib.get_logger()


class ResourceTagsService(object):
    """The service methods for managing a specific resource tag."""

    @inject
    def __init__(
        self,
        resource_type: str,
        resource_id_service: ClassAssistedBuilder[ResourceIdService],
        tag_id_service: TagIdService,
    ) -> None:
        """Initialize the resource tags service.

        All arguments are provided via dependency injection.

        Args:
            resource_id_service: A ResourceIdService object.
            tag_id_service: A TagIdService object.
        """
        self._resource_id_service = resource_id_service.build(
            resource_type=resource_type
        )
        self._tag_id_service = tag_id_service

    def get(
        self,
        resource_id: int,
        **kwargs,
    ) -> list[models.Tag]:
        """Fetch tags for a resource by resource id.

        Args:
            resource_id: The unique id of the resource.

        Returns:
            The list of tags if the resource is found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the resource is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get tags by resource id", resource_id=resource_id)

        resource = self._resource_id_service.get(resource_id, log=log)

        return list(resource.tags)

    def append(
        self,
        resource_id: int,
        tag_ids: list[int],
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> list[models.Tag]:
        """Append one or more Tags

        Args:
            resource_id: The unique id of the resource.
            tag_ids: The list of tag ids to append.
            error_if_not_found: If True, raise an error if the resource is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated tag resource object.

        Raises:
            EntityDoesNotExistError: If the resource is not found and
                `error_if_not_found` is True.
            TagDoesNotExistError: If one or more tags are not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        resource = self._resource_id_service.get(resource_id, log=log)

        tag_ids = list(set(tag_ids))
        existing_tag_ids = set(tag.tag_id for tag in resource.tags)
        tags = [
            self._tag_id_service.get(
                tag_id, error_if_not_found=error_if_not_found, log=log
            )
            for tag_id in tag_ids
            if tag_id not in existing_tag_ids
        ]

        resource.tags.extend(cast(list[models.Tag], tags))

        if commit:
            db.session.commit()
            log.debug("Tags appended successfully", tag_ids=tag_ids)

        return resource.tags

    def modify(
        self,
        resource_id: int,
        tag_ids: list[int],
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> list[models.Tag]:
        """Append one or more Tags

        Args:
            resource_id: The unique id of the resource.
            tag_ids: The list of tag ids to append.
            error_if_not_found: If True, raise an error if the resource is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated tag resource object.

        Raises:
            EntityDoesNotExistError: If the resource is not found and
                `error_if_not_found` is True.
            EntityDoesNotExistError: If one or more tags are not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        resource = self._resource_id_service.get(resource_id, log=log)

        tags = [
            self._tag_id_service.get(
                tag_id, error_if_not_found=error_if_not_found, log=log
            )
            for tag_id in tag_ids
        ]

        resource.tags = cast(list[models.Tag], tags)

        if commit:
            db.session.commit()
            log.debug("Tags updated successfully", tag_ids=tag_ids)

        return resource.tags

    def delete(self, resource_id: int, **kwargs) -> dict[str, Any]:
        """Remove tags from a resource.

        Args:
            resource_id: The unique id of the resource.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        resource = self._resource_id_service.get(resource_id, log=log)
        tag_ids = [tag.tag_id for tag in resource.tags]
        resource.tags = []

        db.session.commit()
        log.debug("Tags removed", resource_id=resource_id, tag_ids=tag_ids)

        return {"status": "Success", "id": tag_ids}


class ResourceTagsIdService(object):
    """The service methods for managing a specific resource tag."""

    @inject
    def __init__(
        self,
        resource_type: str,
        resource_id_service: ClassAssistedBuilder[ResourceIdService],
        tag_id_service: TagIdService,
    ) -> None:
        """Initialize the resource tags service.

        All arguments are provided via dependency injection.

        Args:
            resource_id_service: A ResourceIdService object.
            tag_id_service: A TagIdService object.
        """
        self._resource_id_service = resource_id_service.build(
            resource_type=resource_type
        )
        self._tag_id_service = tag_id_service

    def delete(self, resource_id: int, tag_id, **kwargs) -> dict[str, Any]:
        """Remove a tag from a resource.

        Args:
            resource_id: The unique id of the resource.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        resource = self._resource_id_service.get(
            resource_id, error_if_not_found=True, log=log
        )

        current_tags = resource.tags
        tag_exists = tag_id in {tag.tag_id for tag in current_tags}
        if not tag_exists:
            raise EntityDoesNotExistError(
                TAG_RESOURCE_TYPE, resource_id=resource_id, tag_id=tag_id
            )

        resource.tags = [tag for tag in current_tags if tag.tag_id != tag_id]

        db.session.commit()
        log.debug("Tag removed", resource_id=resource_id, tag_id=tag_id)

        return {"status": "Success", "id": [tag_id]}


class ResourceIdService(object):
    """The service methods for accessing a specific resource."""

    @inject
    def __init__(self, resource_type: str):
        self._resource_type = resource_type

    def get(self, resource_id: int, **kwargs) -> models.Resource:
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug(f"Retrieving {self._resource_type} resource", resource_id=resource_id)

        stmt = select(models.Resource).filter_by(
            resource_id=resource_id, resource_type=self._resource_type, is_deleted=False
        )
        resource = db.session.scalar(stmt)

        if resource is None:
            raise EntityDoesNotExistError(self._resource_type, resource_id=resource_id)

        return resource
